//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: arm_itf.cxx,v 1.3 2004/10/21 23:32:40 schaum Exp $
//----------------------------------------------------------------

#define COSIM_STUB

#include "armcosim.h" // holds list of simulators
#include "arm_itf.h"
#include "external_mem.h"

#include "external_dev.h" // arm coprocessor

// ---- 
map <int, aipblock *>    devicemap;     // memory-mapped interfaces
map <int, emu_device *> cpdevicemap;   // coprocessor interfaces

// ---- ARMSIMCPSINK
// ipblock v(out din  : ns(32);
//           in  dout : ns(32);
//           out adr  : ns(32);
//           out req  : ns(1);
//           in  ack  : ns(1);
//           out rd   : ns(1)) {
//   iptype "armsimcp";
//   ipparm "device=1";
// };

class cpdevice : public emu_device {
  armsimcp *gezelinterface;
public:
  cpdevice(armsimcp *g) : emu_device(), gezelinterface(g) {}
  ~cpdevice() {}
  bool write(dev_data_t data, dev_addr_t addr) {
    gezelinterface->ioval[0]->assignulong((long) data);
    gezelinterface->ioval[2]->assignulong((long) addr);

    gezelinterface->ioval[5]->assignulong((long) 1);
    gezelinterface->touch();

    gezelinterface->requestcounter = 2;
    gezelinterface->ioval[3]->assignulong((long) 1);
    return gezelinterface->ioval[4]->toulong();
  }
  
  bool read(dev_data_t& data, dev_addr_t addr) {
    gezelinterface->ioval[1]->assignulong((long) addr);
    data = gezelinterface->ioval[1]->toulong();

    gezelinterface->ioval[5]->assignulong((long) 0);
    gezelinterface->touch();

    gezelinterface->requestcounter = 2;
    gezelinterface->ioval[3]->assignulong((long) 1);
    return gezelinterface->ioval[4]->toulong();
  }
};

armsimcp::armsimcp(char *name) : aipblock(name) {
  hook           = 0;
  requestcounter = 0;
}

void armsimcp::setparm(char *_name) {
  gval *v = make_gval(32,0);
  
  hook = glbSims[0];

  if (matchparm(_name, "device", *v)) {
    cerr << "armsimcp: set device " << v->toulong() << "\n";

    if (v->toulong() == 0) {
      cerr << "armsimcp: device id 0 is RESERVED - do not use.\n";
    }
    if (cpdevicemap.find(v->toulong()) != cpdevicemap.end()) {
      cerr << "armsimcp: duplicate device ID found\n";
      // already exists
    } else {
      cpdevice *C = new cpdevice(this);
      cpdevicemap[v->toulong()] = C;
      hook->dev_master->register_device(v->toulong(), C);
    } 
 
    return;
  }

  cerr << "armsimcp: parameter " << _name << " not understood\n";
}

bool armsimcp::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isoutput(d) && isname(tname, "din"));
    break;
  case 1:
    return (isinput(d)  && isname(tname, "dout"));
    break;
  case 2:
    return (isoutput(d) && isname(tname, "adr"));
    break;
  case 3:
    return (isoutput(d) && isname(tname, "req"));
    break;
  case 4:
    return (isinput(d)  && isname(tname, "ack"));
    break;
  case 5:
    return (isoutput(d) && isname(tname, "rd"));
    break;
  }
  return false;
}

void armsimcp::run() {
  if (requestcounter) {
    if (requestcounter < 2)
      ioval[3]->assignulong((long) 0);
    requestcounter--;
  }
}

bool armsimcp::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool armsimcp::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void armsimcp::touch() {
  interfacewritten = true;
}


// ---- ARMSIMSINK --- (memory-mapped HWSW cosimulation)

armsimsink::armsimsink(char *name) : aipblock(name) {
  hook = 0;
}

void armsimsink::setparm(char *_name) {
  gval *v = make_gval(32,0);
  
  hook = glbSims[0];

  if (matchparm(_name, "address", *v)) {
    cerr << "armsimsink: set address " << v->toulong() << "\n";
    deviceid = hook->mem->register_addr(v->toulong()); 
    devicemap[deviceid] = this;
    return;
  }
  cerr << "armsimsink: parameter " << _name << " not understood\n";
}

void armsimsink::run() {
  //  cout << "armsimsink value " << ioval[0].toulong() << "\n";
}

bool armsimsink::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isinput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool armsimsink::cannotSleepTest() {
  return false;
}

// ---- ARMSIMPROBE ---
#include "emumem.h"

armsimprobe::armsimprobe(char *name) : aipblock(name) {
  hook             = 0;
  probedblock      = 0;
  interfacewritten = false;
  runonce          = false;
  // make a 'virtual output'
  ioval.push_back(make_gval(32, 0));
}

void armsimprobe::setparm(char *_name) {
  gval *v = make_gval(32,0);
  char *blockname;

  hook = glbSims[0];

  if (matchparm(_name, "probe", *v)) {
    cerr << "armsimprobe: set address " << v->toulong() << "\n";
    deviceid = hook->mem->register_addr(v->toulong()); 
    devicemap[deviceid] = this;
    return;
  } else if ((blockname = matchparm(_name, "block"))) {
    vector<aipblock *>::iterator p;
    for (p = glbIpblocks.begin(); p != glbIpblocks.end(); p++) {
      if ((*p)->blockisname(blockname)) {
	probedblock = *p;
      }
    }
    if (probedblock == 0) {
      cerr << "armsimprobe: block name " << blockname << " not found\n";
    }
    return;
  }
  cerr << "armsimprobe: parameter " << _name << " not understood\n";
}

void armsimprobe::run() {
  if (runonce) {
    runonce = false;
    char buf[255];
    unsigned memloc = ioval[0]->toulong();
    char *ptr = buf;
    while ((*ptr++ = hook->mem->read_byte((target_addr_t) memloc++))) {}
    *ptr = 0;
    if (probedblock) {
      probedblock->probe(buf);
    } else {
      cerr << "armsimprobe: missing block to probe\n";
    }
  }
}

bool armsimprobe::checkterminal(int n, char *tname, aipblock::iodir d) {
  return false;
}

bool armsimprobe::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool armsimprobe::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void armsimprobe::touch() {
  interfacewritten = true;
  runonce = true;
}


// ---- ARMSIMSOURCE --- (memory-mapped HWSW cosimulation)

armsimsource::armsimsource(char *name) : 
  aipblock(name), 
  interfacewritten(false) {
  hook = 0;
}

void armsimsource::setparm(char *_name) {
  gval *v = make_gval(32,0);

  hook = glbSims[0];

  if (matchparm(_name, "address", *v)) {
    cerr << "armsimsource: set address " << v->toulong() << "\n";
    deviceid = hook->mem->register_addr(v->toulong()); 
    devicemap[deviceid] = this;
  }
}

void armsimsource::run() {
  //  cout << "armsimsink value " << ioval[0].toulong() << "\n";
}

bool armsimsource::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isoutput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool armsimsource::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool armsimsource::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void armsimsource::touch() {
  interfacewritten = true;
}

//----- MEMORY ACCESS INTERCEPT FUNCTIONS

byte_t     ext_read_byte     (int dev) {
  return devicemap[dev]->ioval[0]->toulong();
}

halfword_t ext_read_half_word(int dev) {
  return devicemap[dev]->ioval[0]->toulong();
}

word_t     ext_read_word     (int dev) {
  return devicemap[dev]->ioval[0]->toulong();
}

dword_t    ext_read_dword    (int dev) {
  return devicemap[dev]->ioval[0]->toulong();
}

void ext_write_byte     (int dev, byte_t value) {
  devicemap[dev]->ioval[0]->assignulong((long) value);
  devicemap[dev]->touch();
}

void ext_write_half_word(int dev, halfword_t value) {
  devicemap[dev]->ioval[0]->assignulong((long) value);
  devicemap[dev]->touch();
}

void ext_write_word     (int dev, word_t  value  ) {
  devicemap[dev]->ioval[0]->assignulong((long) value);
  devicemap[dev]->touch();
}

void ext_write_dword    (int dev, dword_t value  ) {
  devicemap[dev]->ioval[0]->assignulong((long) value);
  devicemap[dev]->touch();
}



