//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: armcp_itf.cxx,v 1.1 2004/07/21 23:19:37 schaum Exp $
//--------------------------------------------------------------
#define COSIM_STUB

#include "armzilla.h" // holds list of simulators
#include "armcp_itf.h"
#include "sysconparse.h" // holds list of processor names
#include "external_dev.h" // arm coprocessor

// ---- 
map <int, emu_device *> cpdevicemap;   // coprocessor interfaces


// ipblock v(out din  : ns(32);
//           in  dout : ns(32);
//           out adr  : ns(32);
//           out req  : ns(1);
//           in  ack  : ns(1);
//           out rd   : ns(1)) {
//   iptype "armsimcp";
//   ipparm "device=1";
// };

class cpdevice : public emu_device {
  armzillacp *gezelinterface;
public:
  cpdevice(armzillacp *g) : emu_device(), gezelinterface(g) {}
  ~cpdevice() {}
  bool write(dev_data_t data, dev_addr_t addr) {
    gezelinterface->ioval[0]->assignulong((long) data);
    gezelinterface->ioval[2]->assignulong((long) addr);

    gezelinterface->ioval[5]->assignulong((long) 1);
    gezelinterface->touch();

    gezelinterface->requestcounter = 2;
    gezelinterface->ioval[3]->assignulong((long) 1);
    return gezelinterface->ioval[4]->toulong();
  }
  
  bool read(dev_data_t& data, dev_addr_t addr) {
    gezelinterface->ioval[1]->assignulong((long) addr);
    data = gezelinterface->ioval[1]->toulong();

    gezelinterface->ioval[5]->assignulong((long) 0);
    gezelinterface->touch();

    gezelinterface->requestcounter = 2;
    gezelinterface->ioval[3]->assignulong((long) 1);
    return gezelinterface->ioval[4]->toulong();
  }
};

armzillacp::armzillacp(char *name) : aipblock(name) {
  hook           = 0;
  requestcounter = 0;
}

void armzillacp::setparm(char *_name) {
  gval *v = make_gval(32,0);
  char *proc;
  
  if ((proc = matchparm(_name, "process"))) {
    cerr << "armzillacp: set processor " << proc << "\n";
    int idx = find_proc_index(proc);
    if (idx < 0) {
      cerr << "armzillacp: processor " << proc << " does not exist\n";
      exit(0);
    }
    hook = glbSims[idx];
    return;
  }
  if (matchparm(_name, "device", *v)) {
    if (! hook) {
      cerr << "armzillacp: use address parm only after selecting a processor\n";
      exit(0);
    }
    cerr << "armzillacp: set device " << v->toulong() << "\n";

    if (v->toulong() == 0) {
      cerr << "armzillacp: device id 0 is RESERVED - do not use.\n";
    }
    if (cpdevicemap.find(v->toulong()) != cpdevicemap.end()) {
      cerr << "armzillacp: duplicate device ID found\n";
      // already exists
    } else {
      cpdevice *C = new cpdevice(this);
      cpdevicemap[v->toulong()] = C;
      hook->dev_master->register_device(v->toulong(), C);
    } 
 
    return;
  }

  cerr << "armzillacp: parameter " << _name << " not understood\n";
}

bool armzillacp::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isoutput(d) && isname(tname, "din"));
    break;
  case 1:
    return (isinput(d)  && isname(tname, "dout"));
    break;
  case 2:
    return (isoutput(d) && isname(tname, "adr"));
    break;
  case 3:
    return (isoutput(d) && isname(tname, "req"));
    break;
  case 4:
    return (isinput(d)  && isname(tname, "ack"));
    break;
  case 5:
    return (isoutput(d) && isname(tname, "rd"));
    break;
  }
  return false;
}

void armzillacp::run() {
  if (requestcounter) {
    if (requestcounter < 2)
      ioval[3]->assignulong((long) 0);
    requestcounter--;
  }
}

bool armzillacp::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool armzillacp::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void armzillacp::touch() {
  interfacewritten = true;
}



