//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: blockingdetect.cxx,v 1.1 2004/07/21 23:19:39 schaum Exp $
//--------------------------------------------------------------

#include "blockingdetect.h"
#include "armzilla.h" // holds list of simulators

blockingdetect::blockingdetect(char *name) : aipblock(name) {
  F = 0;
  blocking = false;
  hook = 0;
}

void blockingdetect::setparm(char *_name) {
  char *fl;
  char *proc;
  
  if ((fl = matchparm(_name, "file"))) {
    cerr << "blockingdetect: set file " << fl << "\n";
    F = new ofstream(fl);
    return;
  }

  if ((proc = matchparm(_name, "process"))) {
    cerr << "blockingdetect: set processor " << proc << "\n";
    int idx = find_proc_index(proc);
    if (idx < 0) {
      cerr << "armzillatick: processor " << proc << " does not exist\n";
      exit(0);
    }
    hook = glbSims[idx];
    return;
  }

  cerr << "blockingdetect: parameter " << _name << " not understood\n";
}

void blockingdetect::run() {
  //  cout << "blockingdetect value " << ioval[0].toulong() << "\n";
  if (oldreq != ioval[0]->toulong()) {
    // event on req
    if ( (!blocking) && (ioval[0]->toulong() != ioval[1]->toulong())) {
      *F << "0 " << hook->get_cycle_count() << "\n";
      flush(*F);
      blocking = true;
    }
  }

  if (oldack != ioval[1]->toulong()) {
    // event on ack
    if ( ( blocking) && (ioval[0]->toulong() == ioval[1]->toulong())) {
      *F << "1 " << hook->get_cycle_count() << "\n";
      flush(*F);
      blocking = false;
    }
  }

  oldreq = ioval[0]->toulong();
  oldack = ioval[1]->toulong();
}

bool blockingdetect::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isinput(d) && isname(tname, "req"));
    break;
  case 1:
    return (isinput(d) && isname(tname, "ack"));
    break;
  }
  return false;
}

bool blockingdetect::cannotSleepTest() {
  return false;
}

bool blockingdetect::needsWakeupTest() {
  bool v = false;
  if (oldreq != ioval[0]->toulong())
    v = true;
  if (oldack != ioval[1]->toulong())
    v = true;
  oldreq = ioval[0]->toulong();
  oldack = ioval[1]->toulong();
  return v;
}


