
/* 

--------------------------------------------------------------
 Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
 University of California at Los Angeles
                                                                             
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either version 2 of
 the License, or (at your option) any later version.
                                                                             
 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
                                                                             
 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA

 $Id: sysconnect.ll,v 1.3 2004/07/21 23:19:40 schaum Exp $
--------------------------------------------------------------

*/

%option c++
%option debug

%{

#include <fstream.h>
#include "sysconparse.h"
#include "sysconnect.tab.h"

FlexLexer *currentLexer = 0;
ifstream *instream;

int cpp_openstream(char *f) {
  instream     = new ifstream(f);    
  currentLexer = new connectFlexLexer(instream, &cerr);
  return instream->fail();
}

int openstream(char *f) {
   return cpp_openstream(f);
}

int connectlex() {
  if (currentLexer)
    return currentLexer->yylex();
  return 0;
}

void connecterror(char *err) {
  cerr << err << "\n";
}

void destroy_lexer() {
  if (currentLexer)
     delete currentLexer;
}
  
%}

ws          [ \t]+
D           [0-9]
B           [0-1]
H           [0-9a-fA-F]
id          [a-zA-Z_][a-zA-Z_0-9]*
nl          \n

%%
{nl}        ; /* skip newlines */
{ws}        ; /* skip whitespace */
\/\/.*$     ; /* skip comments */

"core"            { return (connectlval.cmd = CORE);            } 
"args"            { return (connectlval.cmd = ARGS);            }
"exec"            { return (connectlval.cmd = EXEC);            }
"hardware"        { return (connectlval.cmd = HARDWARE);        }
"gezel"           { return (connectlval.cmd = GEZEL);           }
"("               { return (connectlval.cmd = RBRACE_OPEN);     }
")"               { return (connectlval.cmd = RBRACE_CLOSE);    }
"="               { return (connectlval.cmd = EQUAL);           }
","               { return (connectlval.cmd = COMMA);           }
\"(\\.|[^\\"])*\" { connectlval.str = (char *) YYText(); return QSTRING; }

%%

int connectwrap() {

  if (instream) {
    instream->close();
    delete instream;
  }

  return 1;
}

int find_proc_index(char *pn) {
  for (map<int, char *>::iterator n = processor_name.begin();
       n != processor_name.end();
       n++) {
    if (! strcmp(n->second, pn))
      return n->first;
  }
  return -1;
}





