
/* 

--------------------------------------------------------------
 Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
 University of California at Los Angeles
                                                                             
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either version 2 of
 the License, or (at your option) any later version.
                                                                             
 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
                                                                             
 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA

 $Id: sysconnect.y,v 1.1 2003/12/30 22:39:38 schaum Exp $
--------------------------------------------------------------

*/

%{
#define YYDEBUG 1

/* ----- yacc to lex interface & common error variables -- */

#include <iostream>

#ifndef _MSC_VER
using namespace std;
#endif

#include "sysconparse.h"


int numprocs = 0;
map <int, char *>  processor_name;
map <int, char *>  exec_name;
map <int, char *>  args_name;
map <int, char *>  gezel_name;
map <int, char **> processor_argv;
map <int, int>     processor_argc;


char *strip_q(char *s) {
  char buf[255];
  int  len = 255;
  char *s2 = buf;
  while (len-- && *s) {
    if (*s != '\"')
      *s2++ = *s;
    *s++;
  }
  *s2 = 0;
  return strdup(buf);
}

void connecterror(const char *m) {
  cerr << m << "\n";
}



%}

%union {
  char * str;
  int    cmd;
}

%token <cmd> CORE
%token <cmd> ARGS
%token <cmd> EXEC
%token <cmd> HARDWARE
%token <cmd> GEZEL
%token <cmd> RBRACE_OPEN
%token <cmd> RBRACE_CLOSE
%token <cmd> COMMA
%token <str> QSTRING
%token <cmd> EQUAL
%start connect_description

%%

connect_description: connect_element
                   | connect_description {numprocs++} connect_element
                   ;

connect_element:   core_element
                 | gezel_element
                 ;

core_element: RBRACE_OPEN core_spec COMMA exec_spec RBRACE_CLOSE {}
            | RBRACE_OPEN core_spec COMMA exec_spec COMMA args_spec RBRACE_CLOSE
              {};

gezel_element: RBRACE_OPEN gezel_spec RBRACE_CLOSE {};

core_spec: CORE EQUAL QSTRING {
  processor_name[numprocs] = strip_q($3); 
  processor_argv[numprocs] = new (char *)[MAXARGS];
  (processor_argv[numprocs])[0] = processor_name[numprocs];  
  processor_argc[numprocs] = 1;  
};

exec_spec: EXEC EQUAL QSTRING {
  exec_name[numprocs] = strip_q($3);
};

args_spec: ARGS EQUAL QSTRING {
  args_name[numprocs] = strip_q($3);
    // identify arguments;
  int n = processor_argc[numprocs];
  char *tok;
  (processor_argv[numprocs])[n++] = strtok(args_name[numprocs], " ");
  while (( tok = strtok(0, " "))) 
    (processor_argv[numprocs])[n++] = tok;
  processor_argc[numprocs] = n;
};

gezel_spec: GEZEL EQUAL QSTRING {
  gezel_name[numprocs] = strip_q($3);
};

%%

int call_sysconnect_parser(char *filename) {
  int retval;

#ifdef YYDEBUG
  yydebug = 0;
#endif

  if (openstream(filename) != 0) {
    printf("File open error\n");
  }

  retval = connectparse();
  destroy_lexer();

  if (retval) {
    printf("Parse Error\n");
  }

  return retval;
}
