//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: gvalint.h,v 1.2 2004/02/05 23:31:37 schaum Exp $
//--------------------------------------------------------------

#ifndef GVALINT_H
#define GVALINT_H

#include <iostream>
#include "gval.h"

using namespace std;

class gvalint : public gval {  
  long value;
  int sign;
  int wordlength;
  int curbase;

  int mask;
  int mask_1;

  void init_mask(unsigned wl);
  void align(const gval &o1, const gval &o2, int &vv1, int &vv2);

 public:
  
  gvalint(unsigned _wordlength, unsigned _sign);
  gvalint(char *v);
  gvalint(unsigned _wordlength, unsigned _sign, char *valuestr);

  void rescale();

  gval & operator = (const gval &  v) ;
  void valuecopy(const gval &  v) ; // like 'operator =' without rescale()
  
  unsigned long toulong() const ;
    signed long toslong() const ; 
  void          assignulong(unsigned long); 
  void          assignslong(signed long); 

  unsigned getwordlength() const ;
  void     setwordlength(unsigned _wordlength) ;
  unsigned getsign() const ;
  void     setbase(unsigned n) ;
  
  const mpzcl &   promote_mpzcl   () const;
  const gvalint & promote_gvalint () const;
  void  copy_mpzcl(const mpzcl &);

  bool operator == (const gval & z2) ;
  bool operator != (const gval & z2) ;
  bool operator <= (const gval & z2) ;
  bool operator >= (const gval & z2) ;
  bool operator <  (const gval & z2) ;
  bool operator >  (const gval & z2) ;
  
  void and_gval   (const gval &o1, const gval &o2) ;
  void ior_gval   (const gval &o1, const gval &o2) ;
  void xor_gval   (const gval &o1, const gval &o2) ;  
  void shr_gval   (const gval &o1, const gval &o2) ;
  void shr_ui_gval(const gval &o1, const unsigned int o2);
  void shl_gval   (const gval &o1, const gval &o2) ;
  void add_gval   (const gval &o1, const gval &o2) ;
  void concat_gval(const gval &o1, const gval &o2);
  void sub_gval   (const gval &o1, const gval &o2) ;
  void mod_gval   (const gval &o1, const gval &o2) ;
  void mul_gval   (const gval &o1, const gval &o2) ;
  void not_gval   (const gval &o1) ;
  void neg_gval   (const gval &o1) ;
  
  void readstream (istream &is) ;
  void writestream(ostream &os) ;

};  

istream & operator >> (istream &is, gvalint &v);
ostream & operator << (ostream &os, gvalint &v);

#endif

