//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: module.cxx,v 1.3 2004/04/18 22:13:41 schaum Exp $
//--------------------------------------------------------------

#include "module.h"
#include <iostream>

moduletable::moduletable() {}

void moduletable::addmodule(symid _mod) {
  mtable.push_back(_mod);
  parentmap[_mod] = NOSYMBOL;
}

void moduletable::defineparentof(symid child, symid parent) {
  parentmap[child] = parent;
}

void moduletable::show() {
  typedef vector<symid>::iterator v_t;
  v_t i;
  cerr << "Module table\n";
  for (i = mtable.begin(); (i < mtable.end()); i++) {
    cerr << (*i) << " parent " << parentmap[*i] << "\n";
  }
}

void moduletable::hierarchysort() {
  typedef vector<symid>::iterator v_t;
  v_t i,j;
  bool swap;

  swap = true;
  while (swap) {
    swap = false;
    for (i = mtable.begin(); (i < mtable.end()) && (swap == false); i++) {
      for (j = i; (j < mtable.end()) && (swap == false); j++) {	
	if (i == j) continue;	
	if (parentmap[(*j)] == (*i)) {
	  symid t = (*j);
	  (*j) = (*i);
	  (*i) = t;
	  swap = true;
	}
      }    
    }
  }
}

symid moduletable::bottomup_iterator_init() {
  hierarchysort();
  mtable_iter = 0;
  symbol_iter = mtable[mtable_iter];
  return symbol_iter;
}

symid moduletable::bottomup_iterator_next() {
  symbol_iter++;
  if (parentmap.find(symbol_iter) != parentmap.end()) {
    // this is a new module. switch to new mtable_iter
    mtable_iter++;
    if (mtable_iter >= mtable.size()) { // EOT
      symbol_iter = (unsigned) NOSYMBOL;
    } else {
      symbol_iter = mtable[mtable_iter];
    }
  }
  return symbol_iter;
}

moduletable glbModuleTable;
