//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: mpzcl.h,v 1.1.1.1 2003/09/25 00:01:51 schaum Exp $
//--------------------------------------------------------------

#ifndef OOGMP_H
#define OOGMP_H

// wrapper class for GNU multiprecision library
#include "gmp.h"
#include <iostream>
#include "gval.h"

using namespace std;

class mpzcl;

class mpzcl_part {
  mpzcl *parent;
  unsigned pos;
  unsigned len;
 public:
  mpzcl_part(mpzcl *_parent, unsigned _pos, unsigned _len);
  mpzcl *  getparent() const;
  unsigned getpos() const;
  unsigned getlen() const;
  mpzcl_part & operator = (const mpzcl &v);
};

class mpzcl : public gval {
 public:
  static int defaultbase;
  enum _mpzclsign {ns=0, tc=1};
 private:

  mpz_t value;
  mpz_t mask;  
  mpz_t mask_1;
  unsigned sign;
  unsigned wordlength;
  unsigned curbase;
  
  void init_mask(unsigned w);
  void rescale();
  mpzcl(unsigned _wordlength, unsigned _sign, mpz_t v);

public:
  
  mpzcl(unsigned _wordlength, unsigned _sign);
  mpzcl(char *bv);
  mpzcl(const mpzcl &v);

  const mpzcl &   promote_mpzcl   () const;
  const gvalint & promote_gvalint () const;
  void  copy_gvalint(const gvalint &);

  // THIS CONSTRUCTOR SUSPICIOUS !!
  mpzcl(unsigned _wordlength, unsigned _sign, char *valuestr, unsigned base=defaultbase);

  unsigned getwordlength() const;
  void setwordlength(unsigned _wordlength);
  unsigned getsign() const { return sign;}

  void setbase(unsigned n);

  unsigned issigned() const {
    return sign;
  }

  virtual ~mpzcl();

  mpzcl & operator = (const mpzcl &  v);
  void valuecopy(const mpzcl &  v); // like 'operator =' without rescale()
  mpzcl & duplicate  (const mpzcl &z1);
  mpzcl & operator = (double          v);
  mpzcl & operator = (signed long int v);

  unsigned long toulong() const;
    signed long toslong() const; 
  void assignulong(unsigned long); 
  void assignslong(signed long); 


  // CREATION OF PART-OF BITVECTORS
  mpzcl_part operator () (unsigned pos, unsigned len = 1);
  mpzcl(mpzcl_part v);

  // ARITHMETIC
  mpzcl operator -  ();
  mpzcl operator +  (const mpzcl &o2);
  mpzcl operator -  (const mpzcl &o2);
  mpzcl operator *  (const mpzcl &o2);
  mpzcl operator /  (const mpzcl &o2);
  mpzcl operator %  (const mpzcl &o2);
  void  operator += (const mpzcl &o2);
  void  operator -= (const mpzcl &o2);
  void  operator *= (const mpzcl &o2);
  void  operator /= (const mpzcl &o2);
  mpzcl abs         (const mpzcl &o2);
  
  // SHIFT
  mpzcl operator << (const unsigned int o2);
  mpzcl operator << (const mpzcl &o2);
  mpzcl operator >> (const unsigned int o2);
  mpzcl operator >> (const mpzcl &      o2);
  void operator <<= (const mpzcl &o2);
  void operator >>= (const mpzcl &o2);

  // LOGICAL
  bool operator == (const mpzcl & z2);
  bool operator != (const mpzcl & z2);
  bool operator <= (const mpzcl & z2);
  bool operator >= (const mpzcl & z2);
  bool operator <  (const mpzcl & z2);
  bool operator >  (const mpzcl & z2);
  bool isidentical  (const mpzcl & z2);

  // BITWISE OPERATORS
  void  copy_bit  (const mpzcl &z2, unsigned to_bitnum, unsigned from_bitnum);
  void  set_slice (const mpzcl &z2, unsigned start);
  mpzcl get_slice (unsigned start, unsigned count);

  mpzcl operator & (const mpzcl & z2);
  mpzcl operator | (const mpzcl & z2);
  mpzcl operator ^ (const mpzcl & z2);
  mpzcl operator ~ ();
  mpzcl get_bit   (unsigned bitnum);
  void  set_bit   (unsigned bitnum);
  void  clr_bit   (unsigned bitnum);

  // non-operator versions (for lhs-by-reference)  
  friend void add_mpzcl   (mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void concat_mpzcl(mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void sub_mpzcl   (mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void mul_mpzcl   (mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void mod_mpzcl   (mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);

  friend void and_mpzcl(mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void ior_mpzcl(mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void xor_mpzcl(mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);

  friend void shl_mpzcl(mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);
  friend void shr_mpzcl(mpzcl &lhs, const mpzcl &o1, const mpzcl &o2);

  friend void not_mpzcl(mpzcl &lhs, const mpzcl &o1);
  friend void neg_mpzcl(mpzcl &lhs, const mpzcl &o1);
  
  friend void slice_mpzcl(mpzcl &lhs, const mpzcl &o1, int idx, int len);

  // I/O
  friend void printvalue(ostream &os, mpzcl &v);
  friend void readvalue(istream &is, mpzcl &v);
  friend void showmask(std::ostream &os, mpzcl &v);

  void readstream (istream &is);
  void writestream(ostream &os);

  gval & operator = (const gval &  v) {
    return operator=(v.promote_mpzcl());
  }
  void valuecopy(const gval &  v) {
    valuecopy(v.promote_mpzcl());
  }
  bool operator == (const gval & z2) {
    return operator ==(z2.promote_mpzcl());
  }
  bool operator != (const gval & z2) {
    return operator !=(z2.promote_mpzcl());
  }
  bool operator <= (const gval & z2) {
    return operator <=(z2.promote_mpzcl());
  }
  bool operator >= (const gval & z2) {
    return operator>=(z2.promote_mpzcl());

  }
  bool operator <  (const gval & z2) {
    return operator<(z2.promote_mpzcl());
  }
  bool operator >  (const gval & z2) {
    return operator>(z2.promote_mpzcl());
  }


  void and_gval  (const gval &o1, const gval &o2) {
    and_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void ior_gval  (const gval &o1, const gval &o2) {
    ior_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void xor_gval  (const gval &o1, const gval &o2)  {
    xor_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  } 
  void shr_gval  (const gval &o1, const gval &o2) {
    shr_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void shl_gval  (const gval &o1, const gval &o2) {
    shl_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void add_gval  (const gval &o1, const gval &o2) {
    add_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void concat_gval(const gval &o1, const gval &o2) {
    concat_mpzcl(*this, 
		 o1.promote_mpzcl(),
		 o2.promote_mpzcl()
		 );
  }
  void sub_gval  (const gval &o1, const gval &o2) {
    sub_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void mod_gval  (const gval &o1, const gval &o2) {
    mod_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void mul_gval  (const gval &o1, const gval &o2) {
    mul_mpzcl(*this, 
	      o1.promote_mpzcl(),
	      o2.promote_mpzcl()
	      );
  }
  void slice_gval(const gval &o1, int idx, int len) {
    slice_mpzcl(*this,
		o1.promote_mpzcl(),
		idx, len);
  }
  void not_gval  (const gval &o1) {
    not_mpzcl(*this, 
	      o1.promote_mpzcl()
	      );
  }
  void neg_gval  (const gval &o1) {
    neg_mpzcl(*this, 
	      o1.promote_mpzcl()
	      );
  }

};

istream & operator >> (istream &is, mpzcl &v);
ostream & operator << (ostream &os, mpzcl &v);

void  mpzcl_init();
void  mpzcl_wrapup();
void *mpzcl_alloc(size_t); 
void *mpzcl_realloc(void *, size_t o, size_t n); 
void  mpzcl_free(void *, size_t);

#endif
