//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: rtprofile.cxx,v 1.3 2004/12/07 00:53:14 schaum Exp $
//--------------------------------------------------------------

#include "rtprofile.h"

rtprofile *glbRTProfile = 0;

rtprofile::rtprofile(bool _upedgeMode, bool _cycleMode) { 
  upedgeMode = _upedgeMode;
  cycleMode  = _cycleMode;
}

void rtprofile::makeop(symid id) {
  //   cerr << "make " << id << "\n";
  if (id == NOSYMBOL) // is an intermediate symbol
    return;
  ops[id] = new rtprofile_op(upedgeMode);
}

void rtprofile::evalop(symid id, gval *v) {
  //    cerr << "evalop " << *v << "\n";
  if (id == NOSYMBOL) // is an intermediate symbol
    return;
  if (ops.find(id) == ops.end()) 
    cerr << "Evalop did not find id " << id << "\n";
  else
    ops[id]->hammingupdate(v);
}

void rtprofile::accumulate() {
  evals.clear();
  toggles.clear();
  // first collect statistics per operation type
  map<symid,  rtprofile_op *>::iterator i;
  for (i = ops.begin(); i != ops.end(); i++) {
    evals  [glbSymboltable.getsymbol(i->first)->type()] += i->second->evals;
    toggles[glbSymboltable.getsymbol(i->first)->type()] += i->second->toggles;
  }
}

void rtprofile::tick(ostream &os, unsigned long n) {
  unsigned totaleval, totaltoggle;
  static unsigned lasteval = 0, lasttoggle = 0;


  if (cycleMode) {

    accumulate();

    totaleval = 0;
    totaltoggle = 0;

    map<symid,  rtprofile_op *>::iterator i;
    for (i = ops.begin(); i != ops.end(); i++) {
      totaleval    += i->second->evals;
      totaltoggle  += i->second->toggles;
    }
    os << "Profile Cycle " << setw(15) << n << ": ";
    os << setw(15) << totaleval - lasteval << " evals, ";
    os << setw(15) << totaltoggle - lasttoggle << " toggles\n";

    lasteval   = totaleval;
    lasttoggle = totaltoggle;
  }
    
}

void rtprofile::report(ostream &os) {

  if (!cycleMode) {

    accumulate();

    // header
    os << setw(15) << "Type";
    os << setw(15) << "Evals";
    os << setw(15) << "Toggles\n";
    
    map<symbol::_symtype, unsigned>::iterator j;
    map<symbol::_symtype, unsigned>::iterator k;
    for (j = evals.begin(), k = toggles.begin();
	 j != evals.end();
	 j++, k++) {
      os << setw(15);
      os << symbol::typestr(k->first);
      os << setw(15);
      os << j->second;
      os << setw(15);
      os << k->second;
      os << "\n";
    }
  }

}
