//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: rtprofile.h,v 1.3 2004/12/07 00:53:15 schaum Exp $
//--------------------------------------------------------------

#ifndef RTPROFILE_H
#define RTPROFILE_H

#include <map>
#include "gval.h"
#include <iomanip>
#include "asymbol.h"
#include <assert.h>

// using namespace std;
// using std::setw;

class rtprofile_op {
  gval *v;
  gval *vand;
  gval *vxor;
  bool upedgeMode;
 protected:
  unsigned evals;
  unsigned toggles;
 public:
  friend class rtprofile;
  rtprofile_op(bool _upedgeMode) {
    v    = make_gval(32, 0);
    vand = make_gval(32,0);
    vxor = make_gval(32,0);
    v->assignulong(0);
    vand->assignulong(0);
    vxor->assignulong(0);
    evals      = 0;
    toggles    = 0;
    upedgeMode = _upedgeMode;
  }
  void hammingupdate(gval *nv) {
    evals++;
    if (upedgeMode) {
      vxor->xor_gval(*v, *nv);
      vand->and_gval(*vxor, *nv);
      toggles += vand->popcount();
    } else {
      toggles += v->hamdist(*nv);
    }
    v->valuecopy(*nv);
  }
};

class rtprofile {
  map<symid,  rtprofile_op *>   ops;

  map<symbol::_symtype, unsigned> evals;
  map<symbol::_symtype, unsigned> toggles;

  bool cycleMode;
  bool upedgeMode;

  void accumulate();

 public:
  rtprofile(bool _upedgeMode, bool _cycleMode);
  void makeop(symid id);
  void evalop(symid id, gval *v);
  void tick(ostream &os, unsigned long n);
  void report(ostream &os);
};

extern rtprofile *glbRTProfile;

#endif
