//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: rttrace.cxx,v 1.4 2004/01/28 02:27:41 schaum Exp $
//--------------------------------------------------------------

#include "rttrace.h"

//-----------------------------------------------------

tracerdumper::tracerdumper(char *file) : os(file), vlen(1024) {
  if (os.fail()) {
    cout << "Error: Unable to open file " << file << " for writing\n";
    exit(0);
  }
  for (unsigned i=0; i<vlen; ++i) {
    buffer.push_back(make_gval("0xdeadbeefdeadbeef"));
  }
  wptr = 0;
}

void tracerdumper::setwl(int wl) {
  for (unsigned i=0; i<vlen; ++i) {
    buffer[i]->setwordlength(wl);
    buffer[i]->setbase(2);         // binary
  }
}

tracerdumper::~tracerdumper() {
  for (unsigned i=0; i<wptr; ++i) { // flush
    os << *buffer[i] << "\n";
  }
  os.close();
}

void tracerdumper::addval(gval *v) {
  *(buffer[wptr++]) = *v;
  //  cerr << wptr << "\n";
  if (wptr >= vlen) { // flush
    for (unsigned i=0; i<vlen; ++i) {
      //      cerr << "flush " << i << "\n";
      os << *buffer[i] << "\n";
    }
    wptr = 0;
  }
}

//-----------------------------------------------------

tracerblock::tracerblock(char *_fname, artsignal *_sig) :
  fname(_fname), sig(_sig), td(_fname) {
  td.setwl(_sig->wlenreq());
}

bool tracerblock::isdupfile(char *_fname) {
  return (strcmp(_fname, fname) == 0);
}

void tracerblock::eval(unsigned long C) {
  td.addval(&(sig->eval(C)) );
}

