//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: vhdout.cxx,v 1.10 2005/01/21 21:58:39 herwin Exp $
//--------------------------------------------------------------

#include "vhdout.h"
#include "rterror.h"

//--------------------------------------------------------------
vhdout::vhdout(string filename) : os(filename.c_str()) {
  //  os.open(filename.c_str());

  indent = 0;

  if (os.fail()) {
      fdlerror(ERROR_CANNOT_OPEN_VHD_FILE, filename.c_str());
  }
}

vhdout::~vhdout() {}

void vhdout::comment(string str) {
  doindent();
  os <<"--"<<str<<"\n";  
}

void vhdout::library(string str) {
  doindent();
  os <<"library "<<str<<";"<<"\n";
}

void vhdout::lib_use(string str) {
  doindent();
  os <<"use "<<str<<";"<<"\n";
}

void vhdout::increaseindent() {
  indent++;
}

void vhdout::decreaseindent() {
  indent--;
}

void vhdout::doindent() {
  for (int i = 0; i < indent; i ++)
	os << "   ";
}

void vhdout::outputline(string str) {
  doindent();
  os << str<<"\n";
}

void vhdout::outputchar(string str) {
  os << str;
}

void vhdout::nextline(int line) {
  for (int i = 0; i < line; i ++)
	os << "\n";
}

void vhdout::beginentity(string name) {
  doindent();
  os <<"entity "<<name<<" is"<<"\n";
  increaseindent();
  doindent();
  os <<"port("<<"\n";
  increaseindent();
  first = true;
}

#include <iostream.h>

void vhdout::dumpportvar(vhdvar* v) {
  //  cerr << "dumpportvar: var name is " <<  v->getportname() << "\n";

  if (!first) {
    os <<";"<<"\n";
  } else first = false;

  doindent();
  os << v->getportname()<<":";
  if (v->isout()) os <<"out ";
  else os <<"in ";
  if (v->getwl() > 1)
    os <<"std_logic_vector("<<v->getwl()-1<<" downto 0)";
  else
    os <<"std_logic";
}

void vhdout::endentity (string name, bool hasclk, string clkname, string rstname) {
  if(hasclk) {
    if(first) {
      doindent();
      os <<rstname<<" : in std_logic;"<<"\n";
      doindent();
      os <<clkname<<" : in std_logic"<<"\n";
    }
    else {
      os <<";"<<"\n";
      doindent();
      os <<rstname<<" : in std_logic;"<<"\n";
      doindent();
      os <<clkname<<" : in std_logic"<<"\n";
    }
  }
  
  os <<"\n";
  decreaseindent();
  doindent();
  os <<");"<<"\n";

  decreaseindent();
  doindent();
  os <<"end "<<name<<";"<<"\n";
}  

void vhdout::begincomponent(string name) {
  doindent();
  os <<"component "<<name<<"\n";
  increaseindent();
  doindent();
  os <<"port("<<"\n";
  increaseindent();
  first = true;
}

void vhdout::endcomponent (string name, bool hasclk, string clkname, string rstname) {
  if(hasclk) {
    if (first) {
       doindent();
       os <<rstname<<" : in std_logic;"<<"\n";
       doindent();
       os <<clkname<<" : in std_logic"<<"\n";
    }
    else {
       os <<";"<<"\n";    
       doindent();
       os <<rstname<<" : in std_logic;"<<"\n";
       doindent();
       os <<clkname<<" : in std_logic"<<"\n";
    }
  }
  decreaseindent();
  doindent();
  os <<");"<<"\n";

  decreaseindent();
  doindent();
  os <<"end component;"<<"\n";
}  

void vhdout::architecture(string type, string name) {
  doindent();
  os <<"architecture "<<type<<" of "<<name<<" is"<<"\n";
}

void vhdout::dumpvardeclare(vhdvar* v) {
  doindent();
  os <<"signal "<< v->getname()<<":";
  if (v->getwl() > 1)
	os <<"std_logic_vector("<<v->getwl()-1<<" downto 0);"<<"\n";
  else
        os <<"std_logic;"<<"\n";
  if (v->isreg()) {
    os <<"signal "<< v->getname()<<"_wire"<<":";
    if (v->getwl() > 1)
	os <<"std_logic_vector("<<v->getwl()-1<<" downto 0);"<<"\n";
    else
        os <<"std_logic;"<<"\n";
  }  
}  

void vhdout::beginstatedeclare() {
  doindent();
  first = true;
  os << "type STATE_TYPE is (";
}

void vhdout::eachstate(string st) {
  if (first) {
	os << st;
	first = false;
  }
  else
	os <<","<<st;
}

void vhdout::endstatedeclare() {
  os <<");"<<"\n";
  doindent();
  os << "signal STATE:STATE_TYPE;"<<"\n";
}

void vhdout::lutdeclare(string name, int wlen,int size) {
  doindent();
  if (size > 1)
    os <<"Type "<<name<<" is Array (Natural range <>) of std_logic_vector("<<(wlen - 1)<<" downto 0);"<<"\n";
}

void vhdout::luttable(string name,string luttypename,int lutwl,vector<gval *> value, int size) {  
  doindent();
  if(size == 1) {
     os <<"constant "<<name<<" : integer := ";
     for (vector<gval *>::iterator i = value.begin(); i != value.end(); ++i) 
	    os <<(*i)<<";"<<"\n";
  }
  else {
     int cnt = 0;
     first = true;
     os <<"constant "<<name<<" : "<<luttypename<<" := (";
     for (vector<gval*>::iterator i = value.begin(); i != value.end(); ++i) {
        cnt++;
 	if(first)  {
	    os <<"conv_std_logic_vector("<<(*i)->toulong()<<","<<lutwl<<")";
	    first = false;
	}
	else if (cnt<5)
	    os <<","<<"conv_std_logic_vector("<<(*i)->toulong()<<","<<lutwl<<")";
        else {
            os <<",\n"<<"conv_std_logic_vector("<<(*i)->toulong()<<","<<lutwl<<")";
            cnt=0;
        }
     }
     os <<");"<<"\n";
  } 
}

void vhdout::dumpbitstream(vhdvar* v) {
  doindent();
  os <<v->getportname()<<" <= "<<v->getbitstream()<<";"<<"\n";
}

void vhdout::outputram(int ramsize, int ramwl, int addrwl,string clkname) {
  doindent();
  os <<"type ram_type is array ("<<ramsize-1<<" downto 0) of std_logic_vector ("<<ramwl-1<<" downto 0);"<<"\n";
  doindent();
  os << "signal RAM : ram_type;"<<"\n";
  doindent();
  os << "signal sig_1 : std_logic_vector("<<addrwl-1<<" downto 0);"<<"\n";
  doindent();
  os << "begin"<<"\n";
  doindent();
  os << "process("<<clkname<<")"<<"\n";
  doindent();
  os << "begin"<<"\n";
  increaseindent();
  doindent();
  os << "if ("<<clkname<<"'event and "<<clkname<<" = '1') then "<<"\n";
  increaseindent();
  doindent();
  os << "if ( wr = '1' ) then "<<"\n";
  increaseindent();
  doindent();
  os << "RAM(conv_integer(unsigned(address))) <= idata;"<<"\n";
  decreaseindent();
  doindent();
  os << "end if; "<<"\n";
  doindent();
  os << "sig_1 <= address;"<<"\n";
  decreaseindent();
  doindent();
  os << "end if; "<<"\n";
  decreaseindent();
  doindent();
  os << "end process; "<<"\n";
  doindent();
  os << "odata <= RAM(conv_integer(unsigned(sig_1)));"<<"\n";
}

void vhdout::beginportmap(string name,string parent) {
  doindent();
  os <<"label_"<<name<<" : "<<parent<<" port map ("<<"\n";
  increaseindent();
  increaseindent();
  first = true;
}

void vhdout::dumpmapvar(vhdvar* v, vhdvar* t) {
  if (!first) {
    os <<","<<"\n";    
  }
  else
    first = false;

  doindent();
  if(t->getwl() > v->getwl()) {
    os << v->getportname()<<" => "<<t->getportname()<<"("<<v->getwl()-1<<" downto 0)";
    cout<<"VHD WARNING: MISMATCH IN WORDLENGHT IN PORTMAP.."<<v->getname()<<" -> "<<t->getname()<<"\n";
  }
  else
    os << v->getportname()<<" => "<<t->getportname();
}

void vhdout::endportmap (bool hasclk, string clkname, string rstname) {
  if(hasclk) {
    if (first) {
       doindent();
       os <<rstname<<" => "<<rstname<<","<<"\n";
       doindent();
       os <<clkname<<" => "<<clkname<<"\n";
    }
    else {
       os <<","<<"\n";    
       doindent();
       os <<rstname<<" => "<<rstname<<","<<"\n";
       doindent();
       os <<clkname<<" => "<<clkname<<"\n";
    }
  }
  decreaseindent();
  doindent();
  os <<");"<<"\n";
  decreaseindent();
}  

void vhdout::begin() {
  doindent();
  os <<"begin"<<"\n";
}

void vhdout::process(string label, string sense) {
  doindent();
  os <<label<<": process ("<<sense<<")"<<"\n";
}

void vhdout::ifresethigh(string rstname, int resethighval) {
  doindent();
  os << "if ("<<rstname<<" = '" << resethighval << "') then"<<"\n";
}

void vhdout::elseclk(string clkname) {
  doindent();
  os << "elsif "<<clkname<<"' event and "<<clkname<<" = '1' then"<<"\n";
}

void vhdout::ifclk(string clkname) {
  doindent();
  os << "if "<<clkname<<"' event and "<<clkname<<" = '1' then"<<endl;
}

void vhdout::defaultzero(vhdvar* v, bool zero) {
  doindent();
  if (v->issetdefault()) {
   if (zero) {
     if (v->getwl() > 1)
	os << v->getportname()<<" <= (others=>'0');"<<"\n";
     else
	os << v->getportname()<<" <= '0';"<<"\n";
   }
   else
	os << v->getportname()<<" <= "<<v->getname()<<";"<<"\n";
  }
}

void vhdout::endif() {
  doindent();
  os <<"end if;"<<"\n";
}

void vhdout::endprocess(string label) {
  doindent();
  os <<"end process "<<label<<";"<<"\n";
}

void vhdout::casestr(string cmd) {
  doindent();
  os <<"case "<<cmd<<" is"<<"\n";
}

void vhdout::endcase() {
  doindent();
  os <<"end case;"<<"\n";
}

void vhdout::when(string c) {
  doindent();
  os << "when "<<c<<" =>"<<"\n";
}

void vhdout::binoparith(string op, int sign, int wl, string nm, bool lconst, int lwl, string lnm,int lsign,bool rconst, int rwl, string rnm,int rsign) {
  doindent();
  os <<nm<<" <= ";
  
  if (op == " & ") {
     os <<lnm<<op<<rnm<<";"<<endl;
     return;
  }
  
  if (op == " * ") 
     if(wl > 1)
	os << "conv_std_logic_vector(";
     else
	cout<<"VHD ERROR: CANNOT SET * TO LOGIC"<<"\n";

  if ( sign ) 
     os <<"signed(";
  else
     os <<"unsigned(";
  if (lconst)
     if(wl > 1)
	os <<"conv_std_logic_vector("<<lnm<<","<<wl<<"))";
     else 
	os <<"'"<<lnm<<"')"; 
  else if (lwl == wl)
     os <<lnm<<")";
  else
     if(wl > 1)
	if (lwl > wl)
     	   os <<lnm<<"("<<(wl-1)<<" downto 0))";
	else {
	   if(lwl == 1)
		if (lsign)
		   os <<"logic_sign_ext("<<lnm<<","<<wl<<"))";
	        else
		   os <<"logic_zero_ext("<<lnm<<","<<wl<<"))";
	   else {
                os <<"conv_std_logic_vector(";
		if (lsign)
		   os <<"signed("<<lnm<<"),"<<wl<<"))";
	        else
		   os <<"unsigned("<<lnm<<"),"<<wl<<"))";
	   }
	}
     else
	os <<lnm<<"(0))";
  os << op;

  if ( sign ) 
     os <<"signed(";
  else
     os <<"unsigned(";
  if (rconst)
     if(wl > 1)
	os <<"conv_std_logic_vector("<<rnm<<","<<wl<<"))"; 
     else
	os <<"'"<<rnm<<"')"; 
  else if (rwl == wl)
     os <<rnm<<")";
  else
     if(wl > 1)
	if (rwl > wl)
     	   os <<rnm<<"("<<(wl-1)<<" downto 0))";
	else {
	   if(rwl == 1)
		if (rsign)
		   os <<"logic_sign_ext("<<rnm<<","<<wl<<"))";
	        else
		   os <<"logic_zero_ext("<<rnm<<","<<wl<<"))";
	   else {
                os <<"conv_std_logic_vector(";
		if (rsign)
		   os <<"signed("<<rnm<<"),"<<wl<<"))";
	        else
		   os <<"unsigned("<<rnm<<"),"<<wl<<"))";
		}
	}
     else
	os <<rnm<<"(0))";
  if (op == " * ") 
	os << ","<<wl<<");"<<"\n";
  else
	os <<";"<<"\n";
}

void vhdout::binoparithlogic(string op, int sign, int wl, string nm, bool lconst, int lwl, string lnm,int lsign,bool rconst, int rwl, string rnm,int rsign) {
  doindent();
  os <<nm<<" <= ";
  if (lconst)
     if(wl > 1)
	os <<"conv_std_logic_vector("<<lnm<<","<<wl<<")";
     else
	os <<"'"<<lnm<<"';"; 
  else if (lwl == wl)
     os <<lnm;
  else
     if(wl > 1)
	if (lwl > wl)
     	   os <<lnm<<"("<<(wl-1)<<" downto 0)";
	else {
	   if(lwl == 1)
		if (lsign)
		   os <<"logic_sign_ext("<<lnm<<","<<wl<<")";
	        else
		   os <<"logic_zero_ext("<<lnm<<","<<wl<<")";
	   else {
                os <<"conv_std_logic_vector(";
		if (lsign)
		   os <<"signed("<<lnm<<"),"<<wl<<")";
	        else
		   os <<"unsigned("<<lnm<<"),"<<wl<<")";
	   }
	}
     else
	os <<lnm<<"(0)";

  os << op;

  if (rconst)
     if(wl > 1)
	os <<"conv_std_logic_vector("<<rnm<<","<<wl<<");";
     else
	os <<"'"<<rnm<<"';"; 
  else if (rwl == wl)
     os <<rnm<<";";
  else
     if(wl > 1)
	if (rwl > wl)
     	   os <<rnm<<"("<<(wl-1)<<" downto 0);";
	else {
	   if(rwl == 1)
		if (rsign)
		   os <<"logic_sign_ext("<<rnm<<","<<wl<<");";
	        else
		   os <<"logic_zero_ext("<<rnm<<","<<wl<<");";
	   else {
                os <<"conv_std_logic_vector(";
		if (rsign)
		   os <<"signed("<<rnm<<"),"<<wl<<");";
	        else
		   os <<"unsigned("<<rnm<<"),"<<wl<<");";
	   }
       }
     else 
	os <<rnm<<"(0);";
  os <<"\n";
}

void vhdout::binopshift(string op, int sign, int wl, string nm, bool lconst, int lwl, string lnm,int lsign,bool rconst, int rwl, string rnm) {
  doindent();
  os <<nm<<" <= ";
  
  
  if (op == " << ") {
         	if (lwl > 1) {		  
		  if ( sign ) 
		     os <<"conv_std_logic_vector(shl(conv_signed(";
		  else	
		     os <<"conv_std_logic_vector(shl(conv_unsigned(";
		  if (lsign)
		     os <<"signed(";
		  else
		     os <<"unsigned(";
		  if (rconst) 		
		    os <<lnm<<"),"<<wl<<"),conv_unsigned("<<rnm<<","<<wl<<")),"<<wl<<");"<<"\n";
                  else
		    os <<lnm<<"),"<<wl<<"),unsigned("<<rnm<<")),"<<wl<<");"<<"\n";
	        }
		else {
		  if ( sign ) 
		     os <<"conv_std_logic_vector(shl(signed(";
		  else	
		     os <<"conv_std_logic_vector(shl(unsigned(";
                  if (lsign)
		     os << "logic_sign_ext("<<lnm<<","<<wl<<")),";
		  else
		     os << "logic_zero_ext("<<lnm<<","<<wl<<")),";
		  if (rconst) 		
		    os <<"conv_unsigned("<<rnm<<","<<wl<<")),"<<wl<<");"<<"\n";
                  else
		    os <<"unsigned("<<rnm<<")),"<<wl<<");"<<"\n";
		}
  }
  else {
         	if (lwl > 1) {		  
		  if ( sign ) 
		     os <<"conv_std_logic_vector(shr(signed(";
		  else	
		     os <<"conv_std_logic_vector(shr(unsigned(";
		  if (rconst) 		
		    os <<lnm<<"),"<<"conv_unsigned("<<rnm<<","<<wl<<")),"<<wl<<");"<<"\n";
                  else
		    os <<lnm<<"),"<<"unsigned("<<rnm<<")),"<<wl<<");"<<"\n";
                }  
		else {
		  if ( sign ) 
		     os <<"conv_std_logic_vector(shr(signed(";
		  else	
		     os <<"conv_std_logic_vector(shr(unsigned(";
		  if (rconst) 		
		    os <<"conv_std_logic_vector("<<lnm<<","<<wl<<")),"<<"conv_unsigned("<<rnm<<","<<wl<<")),"<<wl<<");"<<"\n";
                  else
		    os <<"conv_std_logic_vector("<<lnm<<","<<wl<<")),"<<"unsigned("<<rnm<<")),"<<wl<<");"<<"\n";
		}
  }

}

void vhdout::binoplogic(string op, string nm, bool lconst, string lnm, int lsign,int lwl, bool rconst, string rnm, int rsign) {
  doindent();
  os<<"if (";
  if (lconst)
     os <<lnm;
  else {
     if(lwl > 1){
        if (lsign)
          os <<"signed("<<lnm<<")";	
	else
	  os <<"unsigned("<<lnm<<")";	
        os << op;
        if (rconst)
           os <<rnm<<") then"<<"\n";
        else
           if (rsign)
    	     os <<"signed("<<rnm<<")"<<") then"<<"\n";	
           else
	     os <<"unsigned("<<rnm<<")"<<") then"<<"\n";	
     }
     else {
	  os <<lnm;
          os << op;
          if (rconst)
            os <<"'"<<rnm<<"') then"<<"\n";
          else
 	    os <<rnm<<") then"<<"\n";	
     }
  }


  increaseindent();
  doindent();
  os <<nm<<" <= "<<"'1';"<<"\n";
  decreaseindent();
  doindent();
  os<<"else"<<"\n";
  increaseindent();
  doindent();
  os <<nm<<" <= "<<"'0';"<<"\n";
  decreaseindent();
  doindent();
  os <<"end if;"<<"\n";
}

void vhdout::dunopnot(string nm,int wl,bool lconst,int lwl, string lnm,int lsign) {
  doindent();
  os <<nm<<" <= "<< " not "; 
  if (lconst)
     if(wl > 1)
	os <<"conv_std_logic_vector("<<lnm<<","<<wl<<");"<<"\n";
     else
	os <<"'"<<lnm<<"';"<<"\n"; 
  else if (wl == lwl)
        os <<lnm<<";"<<"\n";
  else
     if(wl > 1)
	if (lwl > wl)
     	   os <<lnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
 	   if(lwl == 1)
		if (lsign)
          	   os <<"logic_sign_ext("<<lnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<lnm<<","<<wl<<");"<<"\n";	
	   else {
                os <<"conv_std_logic_vector(";
		if (lsign)
		   os <<"signed("<<lnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<lnm<<"),"<<wl<<");"<<"\n";
	   }
	}

     else
	os <<lnm<<"(0);"<<"\n";
}

void vhdout::dunopneg(string nm,int wl,int sign,int lwl, string lnm, bool lconst,int lsign) {
  doindent();
  os <<nm<<" <= "<<" -"; 
  if ( sign ) 
     os <<"signed(";
  else
     os <<"unsigned(";

  if (lconst)
     if(wl > 1)
	os <<"conv_std_logic_vector("<<lnm<<","<<wl<<"));"<<"\n";
     else
	cout<<"VHD ERROR: CANNOT - ONE BIT OBJ"<<"\n";
  else if (wl == lwl)
        os <<lnm<<");"<<"\n";
  else
     if(wl > 1)
	if (lwl > wl)
     	   os <<lnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
	   if(lwl == 1)
		if (lsign)
          	   os <<"logic_sign_ext("<<lnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<lnm<<","<<wl<<");"<<"\n";	
	   else {
                os <<"conv_std_logic_vector(";
		if (lsign)
		   os <<"signed("<<lnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<lnm<<"),"<<wl<<");"<<"\n";
 	   }
	}

     else
	os <<lnm<<"(0));"<<"\n";
}

void vhdout::dunopcast(string nm,int wl,int sign,int lwl, string lnm, bool lconst,int lsign) {
  doindent();
  os <<nm<<" <= "; 

  if (lconst)
     if(wl > 1)
	if (sign)
	   os <<"conv_std_logic_vector(conv_signed("<<lnm<<","<<wl<<"),"<<wl<<");"<<"\n";
	else
	   os <<"conv_std_logic_vector(conv_unsigned("<<lnm<<","<<wl<<"),"<<wl<<");"<<"\n";
     else
	cout<<"VHD ERROR: CANNOT - ONE BIT OBJ"<<"\n";
  else if (wl == lwl)
        if (lwl == 1)
	   os << lnm<<";"<<"\n";
	else if (sign)
	   os <<"conv_std_logic_vector(signed("<<lnm<<"),"<<wl<<");"<<"\n";
	else
	   os <<"conv_std_logic_vector(signed("<<lnm<<"),"<<wl<<");"<<"\n";
  else
     if(wl > 1)
	if (lwl > wl)
     	   os <<lnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
          if (lwl == 1)
	     os <<"conv_std_logic_vector("<<lnm<<","<<wl<<");"<<"\n";
          if (sign)
	     os <<"conv_std_logic_vector(signed("<<lnm<<"),"<<wl<<");"<<"\n";
          else
	     os <<"conv_std_logic_vector(unsigned("<<lnm<<"),"<<wl<<");"<<"\n";
	}
     else
	os <<lnm<<"(0));"<<"\n";
}

void vhdout::teropsel(string nm,int wl,string lnm,int lsign, int lwl, string mnm,bool mconst,int mwl,int msign,bool misbit,string rnm,bool rconst,int rwl,int rsign,bool risbit) {
  doindent();
  if (lwl > 1) {
     if (lsign)
       os <<"if (signed("<<lnm<<") > 0) then"<<"\n";
     else
       os <<"if (unsigned("<<lnm<<") > 0) then"<<"\n";
  }
  else
     os <<"if ("<<lnm<<" = '1') then"<<"\n";


  increaseindent();
  doindent();
  os <<nm<<" <= ";
  if (mconst)
     if(misbit)
	if(wl == mwl)
		os << "to_stdlogicvector("<<mnm<<");"<<"\n";
        else if(wl > mwl)
		os <<"ext(to_stdlogicvector("<<mnm<<"),"<<wl<<");"<<"\n";
	else
		os << "to_stdlogicvector("<<mnm<<")("<<wl-1<<" downto 0);"<<"\n";
     else if(wl > 1)
	os <<"conv_std_logic_vector("<<mnm<<","<<wl<<");"<<"\n";
     else
	os <<"'"<<mnm<<"';"<<"\n"; 
  else if (wl == mwl)
        os <<mnm<<";"<<"\n";
  else
     if(wl > 1)
	if (mwl > wl)
     	   os <<mnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
 	   if(mwl == 1)
		if (msign)
          	   os <<"logic_sign_ext("<<mnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<mnm<<","<<wl<<");"<<"\n";	
	   else {
	        os <<"conv_std_logic_vector(";
		if (msign)
		   os <<"signed("<<mnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<mnm<<"),"<<wl<<");"<<"\n";
	   }
	}
     else
	os <<mnm<<"(0);"<<"\n";
  decreaseindent();
  doindent();
  os <<"else"<<"\n";

  increaseindent();
  doindent();
  os <<nm<<" <= ";

  if (rconst)
     if(risbit)
	if(wl == rwl)
		os << "to_stdlogicvector("<<rnm<<");"<<"\n";
        else if(wl > rwl)
		os <<"ext(to_stdlogicvector("<<rnm<<"),"<<wl<<");"<<"\n";
	else
		os << "to_stdlogicvector("<<rnm<<")("<<wl-1<<" downto 0);"<<"\n";
     else if(wl > 1)
	os <<"conv_std_logic_vector("<<rnm<<","<<wl<<");"<<"\n";
     else
	os <<"'"<<rnm<<"';"<<"\n"; 
  else if (wl == rwl)
        os <<rnm<<";"<<"\n";
  else
     if(wl > 1)
	if (rwl > wl)
     	   os <<rnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
	   if(rwl == 1)
		if (rsign)
          	   os <<"logic_sign_ext("<<rnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<rnm<<","<<wl<<");"<<"\n";	
	   else {
                os <<"conv_std_logic_vector(";  
		if (rsign)
		   os <<"signed("<<rnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<rnm<<"),"<<wl<<");"<<"\n";
	   }
	}
     else
	os <<rnm<<"(0);"<<"\n";
  decreaseindent();
  endif();
}

void vhdout::teropmultsel(bool firstsel,string nm,int wl,string lnm,int lsign, int lwl, string mnm,bool mconst,int mwl,int msign,bool misbit) {
  doindent();
  if (firstsel) os <<"if";
  else		os <<"elsif";

  if (lwl > 1) {
     if (lsign)
       os <<" (signed("<<lnm<<") > 0) then"<<"\n";
     else
       os <<" (unsigned("<<lnm<<") > 0) then"<<"\n";
  }
  else
     os <<" ("<<lnm<<" = '1') then"<<"\n";


  increaseindent();
  doindent();
  os <<nm<<" <= ";
  if (mconst)
     if(misbit)
	if(wl == mwl)
		os << "to_stdlogicvector("<<mnm<<");"<<"\n";
        else if(wl > mwl)
		os <<"ext(to_stdlogicvector("<<mnm<<"),"<<wl<<");"<<"\n";
	else
		os << "to_stdlogicvector("<<mnm<<")("<<wl-1<<" downto 0);"<<"\n";
     else if(wl > 1)
	os <<"conv_std_logic_vector("<<mnm<<","<<wl<<");"<<"\n";
     else
	os <<"'"<<mnm<<"';"<<"\n"; 
  else if (wl == mwl)
        os <<mnm<<";"<<"\n";
  else
     if(wl > 1)
	if (mwl > wl)
     	   os <<mnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
 	   if(mwl == 1)
		if (msign)
          	   os <<"logic_sign_ext("<<mnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<mnm<<","<<wl<<");"<<"\n";	
	   else {
	        os <<"conv_std_logic_vector(";
		if (msign)
		   os <<"signed("<<mnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<mnm<<"),"<<wl<<");"<<"\n";
	   }
	}
     else
	os <<mnm<<"(0);"<<"\n";
  decreaseindent();
}


void vhdout::teropendsel(string nm,int wl,string lnm,int lsign, int lwl, string mnm,bool mconst,int mwl,int msign,bool misbit,string rnm,bool rconst,int rwl,int rsign,bool risbit) {
  doindent();
  if (lwl > 1) {
     if (lsign)
       os <<"elsif (signed("<<lnm<<") > 0) then"<<"\n";
     else
       os <<"elsif (unsigned("<<lnm<<") > 0) then"<<"\n";
  }
  else
     os <<"elsif ("<<lnm<<" = '1') then"<<"\n";


  increaseindent();
  doindent();
  os <<nm<<" <= ";
  if (mconst)
     if(misbit)
	if(wl == mwl)
		os << "to_stdlogicvector("<<mnm<<");"<<"\n";
        else if(wl > mwl)
		os <<"ext(to_stdlogicvector("<<mnm<<"),"<<wl<<");"<<"\n";
	else
		os << "to_stdlogicvector("<<mnm<<")("<<wl-1<<" downto 0);"<<"\n";
     else if(wl > 1)
	os <<"conv_std_logic_vector("<<mnm<<","<<wl<<");"<<"\n";
     else
	os <<"'"<<mnm<<"';"<<"\n"; 
  else if (wl == mwl)
        os <<mnm<<";"<<"\n";
  else
     if(wl > 1)
	if (mwl > wl)
     	   os <<mnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
 	   if(mwl == 1)
		if (msign)
          	   os <<"logic_sign_ext("<<mnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<mnm<<","<<wl<<");"<<"\n";	
	   else {
	        os <<"conv_std_logic_vector(";
		if (msign)
		   os <<"signed("<<mnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<mnm<<"),"<<wl<<");"<<"\n";
	   }
	}
     else
	os <<mnm<<"(0);"<<"\n";
  decreaseindent();
  doindent();
  os <<"else"<<"\n";

  increaseindent();
  doindent();
  os <<nm<<" <= ";

  if (rconst)
     if(risbit)
	if(wl == rwl)
		os << "to_stdlogicvector("<<rnm<<");"<<"\n";
        else if(wl > rwl)
		os <<"ext(to_stdlogicvector("<<rnm<<"),"<<wl<<");"<<"\n";
	else
		os << "to_stdlogicvector("<<rnm<<")("<<wl-1<<" downto 0);"<<"\n";
     else if(wl > 1)
	os <<"conv_std_logic_vector("<<rnm<<","<<wl<<");"<<"\n";
     else
	os <<"'"<<rnm<<"';"<<"\n"; 
  else if (wl == rwl)
        os <<rnm<<";"<<"\n";
  else
     if(wl > 1)
	if (rwl > wl)
     	   os <<rnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
	else {
	   if(rwl == 1)
		if (rsign)
          	   os <<"logic_sign_ext("<<rnm<<","<<wl<<");"<<"\n";	
		else
          	   os <<"logic_zero_ext("<<rnm<<","<<wl<<");"<<"\n";	
	   else {
                os <<"conv_std_logic_vector(";  
		if (rsign)
		   os <<"signed("<<rnm<<"),"<<wl<<");"<<"\n";
	        else
		   os <<"unsigned("<<rnm<<"),"<<wl<<");"<<"\n";
	   }
	}
     else
	os <<rnm<<"(0);"<<"\n";
  decreaseindent();
  endif();
}

void vhdout::assign(string nm,
		    int    wl,
		    bool   lconst,
		    int    lwl,
		    string lnm,
		    int    lsign,
		    bool   lisbit) {
  doindent();
  os <<nm<<" <= ";
  if (lconst)
    if(lisbit)
      if(wl == lwl)
	os << "to_stdlogicvector("<<lnm<<");"<<"\n";
      else if(wl > lwl)
	os <<"ext(to_stdlogicvector("<<lnm<<"),"<<wl<<");"<<"\n";
      else
	os << "to_stdlogicvector("<<lnm<<")("<<wl-1<<" downto 0);"<<"\n";
    else if(wl > 1)
      os <<"conv_std_logic_vector("<<lnm<<","<<wl<<");"<<"\n";
    else
      os <<"'"<<lnm<<"';"<<"\n"; 
  else if (wl == lwl)
    os <<lnm<<";"<<"\n";
  else
    if(wl > 1)
      if (lwl > wl)
	os <<lnm<<"("<<(wl-1)<<" downto 0);"<<"\n";
      else {
	if(lwl == 1)
	  if (lsign)
	    os <<"logic_sign_ext("<<lnm<<","<<wl<<");"<<"\n";	
	  else
	    os <<"logic_zero_ext("<<lnm<<","<<wl<<");"<<"\n";	
	else {
	  os <<"conv_std_logic_vector(";
	  if (lsign)
	    os <<"signed("<<lnm<<"),"<<wl<<");"<<"\n";
	  else
	    os <<"unsigned("<<lnm<<"),"<<wl<<");"<<"\n";
	}
      }
    else
      os <<lnm<<"(0);"<<"\n";
}

void vhdout::lutop(string nm, int wl, string lutnm, int lutsize, string lnm,int lwl,bool lconst) {
  doindent();
  if(lutsize == 1) 
    os << nm<<" <= conv_std_logic_vector("<<lutnm<<","<<wl<<");"<<"\n";  
  else if (lconst)
    os << nm<<" <= "<<lutnm<<"(conv_integer(unsigned(conv_std_logic_vector("<<lnm<<","<<wl<<"))));"<<"\n";
  else	
    os << nm<<" <= "<<lutnm<<"(conv_integer(unsigned("<<lnm<<")));"<<"\n";
}

void vhdout::assignreg(vhdvar* v) {
  doindent();
  os <<v->getname()<<" <= "<<v->getregname()<<";"<<endl;
}

void vhdout::assigndefaultreg(vhdvar* v) {
  doindent();
  os << v->getname()<<"_wire <= "<<v->getname()<<";"<<"\n";
}

void vhdout::close() {
  os.close();
}
