
//-----------------------------------------------------------------------------
// GEZEL/8051 cosimulator
//-----------------------------------------------------------------------------
//
// 8051 simulator based on:
// Roman Lysecky, Tony Givargis, Greg Stitt, Ann Gordon-Ross, and 
// Kris Miller 
//
// Copyright (C) 2001, All Rights Reserved.
//
// 8051 Instruction Set Simulator 
// Version 1.4
//-----------------------------------------------------------------------------

#include <iostream>
#include <csignal>
#include "i8051.h"

#include "rtsim.h"
#include "gval.h"

void SigHandler(int signal);

I8051 glbi8051;

//-----------------------------------------------------------------------------

rtsimgen *glbRTsimgen = 0;

void usage(char *n) {
  fprintf(stderr, 
	  "usage: %s [-g gezel_file] -x hex_file\n"
	  " -d : debug mode\n"
	  " -g : gezel hardware description \n"
	  " -x : i8051 hex file\n"
          " If no gezel file given, will work as i8051 ISS by itself\n"
          "i8051 ISS by the Dalton Project, UCR (http://www.cs.ucr.edu/~dalton/)\n", n);
}

//-----------------------------------------------------------------------------

int main(int argc, char* argv[])  {
  char *gezel_name = 0;
  char *hex_name   = 0;
  int i;
  void (*prev_sigint_handler)(int);

  for(i = 1; i < argc; i++) {
    if(strcmp(argv[i], "-g") == 0) gezel_name = argv[++i]; else
      if(strcmp(argv[i], "-x") == 0) hex_name = argv[++i]; else
	if (strcmp(argv[i], "-d") == 0) glbRTDebug = 1;
  }

  if (argc < 2) {
    usage(argv[0]);
    exit(0);
  }

  if (hex_name == 0) {
    usage(argv[0]);
    exit(0);
  }

  //
  //  if (gezel_name == 0) {
  //    usage(argv[0]);
  //    exit(0);
  //  }

  // load GEZEL program
  glbRTsimgen = new rtsimgen;

  if (gezel_name != 0) {
    printf("Initializing GEZEL processor %s\n", gezel_name);
    fdldebug = 0;
    if (call_parser(gezel_name)) {
      fprintf(stderr, "Could not open %s\n", gezel_name);
      exit(0);
    }
    glbAllowSimulatorSleep = 1;
    glbMPZCustomAllocator = 0;
    glbSymboltable.createcode(glbRTsimgen);
    glbRTsimgen->reset();
  }

  // load & execute i8051 program
  prev_sigint_handler = signal(SIGINT, SigHandler);
  printf("Initializing 8051 processor %s\n", hex_name);
  glbi8051.Simulate(hex_name, "output.txt");
  if (prev_sigint_handler != SIG_ERR) signal(SIGINT, SIG_DFL);
    
  return(0);
}

//-----------------------------------------------------------------------------

void SigHandler(int signal) {
    glbi8051.Stop();
}

//-----------------------------------------------------------------------------
