//--------------------------------------------------------------
// (c) 2004 Patrick Schaumont
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: i8051_itf.cxx,v 1.1 2004/09/16 02:26:00 schaum Exp $
//--------------------------------------------------------------

#include "i8051_itf.h"
#include "rtsim.h"
#include "i8051.h"

extern I8051 glbi8051; // in gezel51.cxx

// ---- I8051SINK --- (memory-mapped HWSW cosimulation)

i8051sink::i8051sink(char *name) : aipblock(name) {}

i8051sink::~i8051sink() {}

void i8051sink::setparm(char *_name) {
  char *port;

  if ((port = matchparm(_name, "port"))) {
    //    port -> contains port
    if (strcmp(port, "P0") == 0) {
      portaddress = glbi8051.P0;
    } else if (strcmp(port, "P1") == 0) {
      portaddress = glbi8051.P1;
    } else if (strcmp(port, "P2") == 0) {
      portaddress = glbi8051.P2;
    } else if (strcmp(port, "P3") == 0) {
      portaddress = glbi8051.P3;
    } else {
      cerr << "i8051sink: unmatched port name (choose one of P0, P1, P2, P3)\n";
      exit(0);
    }
  }

}

void i8051sink::run() {
  glbi8051.writeRAM(portaddress, ioval[0]->toulong());
}

bool i8051sink::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isinput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool i8051sink::cannotSleepTest() {
  return false;
}


// ---- I8051SOURCE --- (memory-mapped HWSW cosimulation)
map <int, aipblock *> i8051dvm; 


i8051source::i8051source(char *name) : aipblock(name), interfacewritten(false) {
  portaddress = 0;
}

i8051source::~i8051source() {
}

void i8051source::setparm(char *_name) {
  char *port;
  if ((port = matchparm(_name, "port"))) {
    //    port -> contains port
    if (strcmp(port, "P0") == 0) {
      portaddress = glbi8051.P0;
      i8051dvm[portaddress] = this;
    } else if (strcmp(port, "P1") == 0) {
      portaddress = glbi8051.P1;
      i8051dvm[portaddress] = this;
    } else if (strcmp(port, "P2") == 0) {
      portaddress = glbi8051.P2;
      i8051dvm[portaddress] = this;
    } else if (strcmp(port, "P3") == 0) {
      portaddress = glbi8051.P3;
      i8051dvm[portaddress] = this;
    } else {
      cerr << "i8051source: unmatched port name (choose one of P0, P1, P2, P3)\n";
      exit(0);
    }
  }
}

void i8051source::run() {
}

bool i8051source::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isoutput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool i8051source::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool i8051source::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void i8051source::touch() {
  interfacewritten = true;
}

//---------------------

void i8051externalwrite(unsigned int a, unsigned char d) {
  if (i8051dvm.find(a) != i8051dvm.end()) {
    i8051dvm[a]->touch();
    i8051dvm[a]->ioval[0]->assignulong((unsigned long) d);
    //    cerr << "extwrite " << a << ": " << (unsigned int) d << "\n";
  }
}


