#include "systemc_itf.h"


map <string, systemcsource *>       systemcsourcelist;
map <string, systemcsink   *>       systemcsinklist;
typedef sc_out < sc_int<32> > *    otptr;
typedef map <systemcsink *, otptr > otmap;
otmap   outputmap;
typedef otmap::iterator omap_iter;

//----------------------------------------------------------------

systemcsource::systemcsource(char *name) : aipblock(name), interfacewritten(false) {}

void systemcsource::setparm(char *_name) {
  char *sysvar;
  if ((sysvar = matchparm(_name, "var"))) {
    cerr << "systemcsource: set variable " << sysvar << "\n";
    systemcsourcelist[string(sysvar)] = this;
  }
}

void systemcsource::run() {}

bool systemcsource::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isoutput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool systemcsource::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool systemcsource::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void systemcsource::touch() {
  interfacewritten = true;
}

//----------------------------------------------------------------

systemcsink::systemcsink(char *name) : aipblock(name) {}

void systemcsink::setparm(char *_name) {
  char *sysvar;
  if ((sysvar = matchparm(_name, "var"))) {
    cerr << "systemcsink: set variable " << sysvar << "\n";
    systemcsinklist[string(sysvar)] = this;
  }
}

void systemcsink::run() {}

bool systemcsink::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isinput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool systemcsink::cannotSleepTest() {
  return false;
}

//----------------------------------------------------------------

gezel_module::gezel_module(sc_module_name scname, char *fdlname) : sc_module(scname) {
  myrtsimgen = 0;
  if (call_parser(fdlname)) {
    cerr << "Could not open GEZEL file " << fdlname << "\n";
    exit(0);
  }
  glbAllowSimulatorSleep = 1;
  glbMPZCustomAllocator = 0;
  myrtsimgen = new rtsimgen;
  glbSymboltable.createcode(myrtsimgen);
  myrtsimgen->reset();    
  SC_METHOD(run);
  sensitive_pos << clk;
}

void gezel_module::run() {
  myrtsimgen->step();
  for (omap_iter i = outputmap.begin(); i != outputmap.end(); i++) {
    i->second->write(i->first->ioval[0]->toslong());
  }
}

//----------------------------------------------------------------

gezel_inport::gezel_inport(sc_module_name scname, char *fdlsigname) : sc_module(scname) {
  
  gzlmodule = 0;
  
  if (systemcsourcelist.find(string(fdlsigname)) != systemcsourcelist.end()) {
    gzlmodule = systemcsourcelist[string(fdlsigname)];
  } else {
    cerr << "Warning - Gezel inport does not include signal " << fdlsigname << "\n";
  }
  
  SC_METHOD(run);
  sensitive << datain;
}

void gezel_inport::run() {
  if (gzlmodule) {
    gzlmodule->ioval[0]->assignslong(datain.read().to_int());
    gzlmodule->touch();
  }
}


//----------------------------------------------------------------

gezel_outport::gezel_outport(sc_module_name scname, char *fdlsigname) : sc_module(scname) {
    gzlmodule = 0;
    if (systemcsinklist.find(string(fdlsigname)) != systemcsinklist.end()) {
      gzlmodule = systemcsinklist[string(fdlsigname)];
    } else {
      cerr << "Warning - Gezel outport does not include signal " << fdlsigname << "\n";
    }
    outputmap[gzlmodule] = &dataout;  
}

