//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// Parts of this code Copyright (C) 2002,2003 Wei Qin
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: armzilla.cxx,v 1.2 2004/09/16 23:41:56 schaum Exp $
//----------------------------------------------------------------

#include "armzilla.h"

static void usage(char *fname) {
  char *cp;
  if ((cp = strrchr(fname, '/'))!=NULL) cp++;
  else cp = fname;
  fprintf(stderr,
	  "**** GEZEL Kernel 1.3 (C) 2001 - 2004 Patrick Schaumont ****\n"
          "**** SimIt-ARM 2.0    (C) 2002 - 2004 Wei Qin           ****\n"
	  "usage : %s [-v] [-f fpe-path] [-d] [-h] <connection file name>\n"
	  "  -v : verbose\n"
	  "  -c : maximum number of clock cycles to simulate\n"
	  "  -f : FPE binary file path\n"
	  "  -d : GEZEL Debug Mode\n"
	  "  -h : print this message and quit\n"
	  "  connection file name : .connect file with ARM ELF32 programs\n", cp);
}
/* ---------------------------------------------------- */

void get_system(char *fname) {

  call_sysconnect_parser(fname);

  if (gezel_name.size() > 1) {
    printf("Only a single gezel file is supported.\n");
    exit(0);
  }

}

void show_system() {
  map<int, char *>::iterator i;
  for (i= processor_name.begin(); i != processor_name.end(); i++) 
    printf("%s\n", i->second);

}

/* ---------------------------------------------------- */

map<int, ARMSIMTYP *> glbSims;
rtsimgen *glbRTsimgen = 0;

/* ---------------------------------------------------- */
static void sigint_handler(int signum) {
  for (map<int, ARMSIMTYP *>::iterator i = glbSims.begin();
       i != glbSims.end();
       i++) {
    i->second->stop(); // Simit-ARM-2.0
  }
}

/* ---------------------------------------------------- */

void init_simulators(bool verbose, char *fpe_name, char *envp[]) {

  map<int, char *>::iterator i;

  for (i= processor_name.begin(); i != processor_name.end(); i++) {
    printf("Initializing StrongARM processor %s\n", i->second);

    ARMSIMTYP *p = new ARMSIMTYP(verbose);
    p->load_fpe(fpe_name);
    p->load_program(exec_name[i->first], 
		    processor_argc[i->first], 
		    processor_argv[i->first], envp);

    p->initialize();

    glbSims[i->first] = p;
  }

  if (gezel_name.size() > 0) {
    printf("Initializing GEZEL processor %s\n", gezel_name.begin()->second);

    /* load FDL program */
    fdldebug = 0;
    if (call_parser(gezel_name.begin()->second)) {
      fprintf(stderr, "Could not open %s\n", gezel_name.begin()->second);
      exit(0);
    }

    ofstream F("fdl.symbols");
    glbSymboltable.show(F);
    F.close();

    glbAllowSimulatorSleep = 1;
    glbMPZCustomAllocator = 0;
    glbRTsimgen = new rtsimgen;
    glbSymboltable.createcode(glbRTsimgen);
    //    mpzcl_init();
    glbRTsimgen->reset();
    
  }

}

/* ---------------------------------------------------- */

void run_simulators(UInt64 max_cnum) {
  UInt64 i;
  bool more = false;

  map <int, ARMSIMTYP *>::iterator proc;

  for (i=0; (i<max_cnum); i++) {    
    more = false;
    for (proc = glbSims.begin(); 
	 proc != glbSims.end(); proc++) {
      if (proc->second->is_running()) { // simit-arm-2.0
	more |= true;
	proc->second->clock_tick();
      }
    }
    if (glbRTsimgen)
      glbRTsimgen->step();
    if (! more)
      break;
  }

  for (proc = glbSims.begin(); proc != glbSims.end(); proc++) 
  proc->second->dump_stats(stdout); // simit-arm-2.0

}

/* ---------------------------------------------------- */

#include <curses.h>

int main(int argc, char *argv[], char *envp[]) {
  int i;
  char *prog_name = NULL;
  char *fpe_name = NULL;
  bool verbose = false;
  UInt64 max_cnum = (UInt64)-1;
  void (*prev_sigint_handler)(int);
  
  for(i = 1; i < argc; i++) {
    if(strcmp(argv[i], "-v") == 0) verbose = true; else
      if(strcmp(argv[i], "-f") == 0) fpe_name = argv[++i]; else
	if(strcmp(argv[i], "-c") == 0) max_cnum = ato_uint64(argv[++i]); else
	  if(strcmp(argv[i], "-h") == 0) {usage(argv[0]); return 0;} else
	    if(strcmp(argv[i], "-d") == 0) glbRTDebug = 1; else
	      if(!prog_name) {
		prog_name = argv[i];
		break;
	      }
  }

  if (argc < 2) {
    usage(argv[0]);
    exit(0);
  }

  get_system(prog_name);

  init_simulators(verbose, fpe_name, envp);
  prev_sigint_handler = signal(SIGINT, sigint_handler);
  run_simulators(max_cnum);
  if (prev_sigint_handler != SIG_ERR) signal(SIGINT, SIG_DFL);

  // for curses
  endwin();
  
  return 0;
}




