// curses interface

#include "cursesip.h"
#include "curses.h"
#include <time.h>


time_t glbTimeStart;
time_t glbTimeNow;
long   glbCPS = 0;

cursesip::cursesip(char *name) : aipblock(name) {
  initscr();
  start_color();
  curs_set(0);
  nodelay(stdscr, TRUE);
  cbreak();
  noecho();
  keypad(stdscr, TRUE);
  init_pair(1, COLOR_WHITE, COLOR_BLUE);
  bkgd(COLOR_PAIR(1));
  wattrset(stdscr, COLOR_PAIR(1) | A_BOLD);

  waitreqlo = 0;

  //----------
  time(&glbTimeStart);

}

cursesip::~cursesip() {
  endwin();
}

void cursesip::setparm(char *_name) {
  cerr << "cursesip: parameter " << _name << " not understood\n";
}

long glbBounces     = 0;  // show this
long glbBounceDelay = 0;
long glbLastBounce  = 0;
long glbBounceTotal = 0;
long glbBounceAvg   = 0;  // show this

void cursesip::showcycle() {
  long k = glbRTCycle;
  int  p = COLS/4 + 5;

  while (k) {
    mvaddch(1, p, (char) ('0' + (k % 10)));
    p--;
    k = k / 10;
  }
  mvaddch(1, p--, ' ');
  mvaddch(1, p--, 'C');

  time(&glbTimeNow);
  glbCPS = (long) (glbRTCycle / difftime(glbTimeNow, glbTimeStart));
  k      = glbCPS;

  p = 3*COLS/4 + 5;
  while (k) {
    mvaddch(1, p, (char) ('0' + (k % 10)));
    p--;
    k = k / 10;
  }
  mvaddch(1, p--, ' ');
  mvaddch(1, p--, 's');
  mvaddch(1, p--, 'p');
  mvaddch(1, p--, 'c');

  k = glbBounces;
  p = COLS/4 + 5;
  while (k) {
    mvaddch(  2, p, (char) ('0' + (k % 10)));
    p--;
    k = k / 10;
  }
  mvaddch(  2, p--, ' ');
  mvaddch(  2, p--, 'b');
  mvaddch(  2, p--, 'N');

  k = glbBounceAvg;
  p = 3 * COLS/4 + 5;
  while (k) {
    mvaddch(  2, p, (char) ('0' + (k % 10)));
    p--;
    k = k / 10;
  }
  mvaddch(  2, p--, ' ');
  mvaddch(  2, p--, 'g');
  mvaddch(  2, p--, 'v');
  mvaddch(  2, p--, 'A');

}

void cursesip::run() {
  unsigned c;
  int x, y, ch;

  //  cerr << ioval[4]->toulong() << "\n";

  if (waitreqlo) {
    if (ioval[4]->toulong()!=0)
      return;
    else {
      waitreqlo = 0;
      return;
    }
  } else {
    if (ioval[4]->toulong()==0)
      return;
    else
      waitreqlo = 1;
  }

  //  cerr << "exec " << ioval[0]->toulong() << " " << ioval[1]->toulong() << " " << ioval[2]->toulong() << "\n";

  c = ioval[0]->toulong();
  switch (c) {
  case 0:
    erase();
    break;
  case 1:
    box(stdscr, 0, 0);
    break;
  case 2:
    x  = (int) ioval[1]->toulong();
    y  = (int) ioval[2]->toulong();
    ch = (int) ioval[3]->toulong();
    //  cerr << 2 << " " << x << " " << y << " " << ch << "\n";
    mvaddch(y, x,  ch);
    break;
  case 3:
    refresh();
    break;
  case 4:
    // increment bounce counter
    glbBounces++;
    glbBounceDelay  = glbRTCycle - glbLastBounce;
    glbBounceTotal += glbBounceDelay;
    if (glbBounces)
      glbBounceAvg    = glbBounceTotal / glbBounces;
    glbLastBounce   = glbRTCycle;
    //-----------
    showcycle();

    break;
  }
}

bool cursesip::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch(n) {
  case 0:
    return (isinput(d) &&  isname(tname, "cmd"));
    break;
  case 1:
    return (isinput(d) &&  isname(tname, "x"));
    break;
  case 2:
    return (isinput(d) &&  isname(tname, "y"));
    break;
  case 3:
    return (isinput(d) &&  isname(tname, "data"));
    break;
  case 4:
    return (isinput(d) &&  isname(tname, "req"));
    break;
  }
  return false;
}

bool cursesip::cannotSleepTest() {
  return false;
}

//-----------------------------------

cursesglobals::cursesglobals( char *name) : aipblock(name) {}

cursesglobals::~cursesglobals() {}

void cursesglobals::setparm(char *_name) {
  cerr << "cursesglobals: parameter " << _name << " not understood\n";
}

void cursesglobals::run() {
  //  cerr << "c " << COLS << "l " << LINES << "\n";
}

bool cursesglobals::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch(n) {
  case 0:
    ioval[0]->assignulong(COLS);
    return (isoutput(d) &&  isname(tname, "cols"));
    break;
  case 1:
    ioval[1]->assignulong(LINES);
    return (isoutput(d) &&  isname(tname, "lines"));
    break;
  }
  return false;
}

bool cursesglobals::cannotSleepTest() {
  return false;
}




                 

                 
