//--------------------------------------------------------------
// Copyright (C) 2004 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: ball.c,v 1.1 2004/03/10 06:28:01 schaum Exp $
//
//--------------------------------------------------------------

#include "field.h"


int field_y;
int field_x;

void get_field_parms() {
  field_y = recv_byte();
  field_x = recv_byte();
}

void listen_collision(int *ball_x, int *ball_y, int *collision_type) {
  while (1) { // round-robin scheme
    if (poll_byte()) {
      *ball_x         = recv_byte();
      *ball_y         = recv_byte();
      *collision_type = recv_byte();
      return;
    } else if (poll_byte_2()) {
      *ball_x         = recv_byte_2();
      *ball_y         = recv_byte_2();
      *collision_type = recv_byte_2();
      return;
    } else if (poll_byte_3()) {
      *ball_x         = recv_byte_3();
      *ball_y         = recv_byte_3();
      *collision_type = recv_byte_3();
      return;
    }
  }
}

void broadcast_movement(int ball_x, int ball_y, int v_x, int v_y, int ct) {
  send_byte(ball_x);
  send_byte(ball_y);
  send_byte(v_x);
  send_byte(v_y);
  send_byte(ct);

  send_byte_2(ball_x);
  send_byte_2(ball_y);
  send_byte_2(v_x);
  send_byte_2(v_y);
  send_byte_2(ct);

  send_byte_3(ball_x);
  send_byte_3(ball_y);
  send_byte_3(v_x);
  send_byte_3(v_y);
  send_byte_3(ct);
}

volatile unsigned int *data_outp;
volatile unsigned int *data_inp ;
volatile unsigned int *data_outp_2;
volatile unsigned int *data_inp_2 ;
volatile unsigned int *data_outp_3;
volatile unsigned int *data_inp_3 ;

int main() {
  int speeds[] = {-2, -1, 1, 2};

  int ball_x, ball_y;
  int v_x, v_y;
  int collision_type;

  data_outp   = (unsigned int *) 0x80000008;
  data_inp    = (unsigned int *) 0x80000000;
  data_outp_2 = (unsigned int *) 0x8000000C;
  data_inp_2  = (unsigned int *) 0x80000010;
  data_outp_3 = (unsigned int *) 0x80000018;
  data_inp_3  = (unsigned int *) 0x80000014;

  // initialization
  get_field_parms();

  ball_x = field_x / 2;
  ball_y = field_y / 2;
  v_x = 2;
  v_y = 1;

  broadcast_movement(ball_x, ball_y, v_x, v_y, BALL_LOST);
  collision_type = BALL_LOST;

  while (1) {

    broadcast_movement( ball_x,  ball_y, v_x, v_y, collision_type);
    listen_collision  (&ball_x, &ball_y, &collision_type);

    switch (collision_type) {
    case BALL_LOST:
      ball_x = field_x / 2;
      ball_y = field_y / 2;
      v_y = speeds[rand() % 4];
      v_x = speeds[rand() % 4];
      break;
    case COLLISION_UP:
    case COLLISION_DOWN:
      v_y = -v_y;
      break;
    case COLLISION_LEFT:
    case COLLISION_RIGHT:
      v_x = -v_x;
      break;
    case COLLISION_UP_LEFT:
    case COLLISION_DOWN_LEFT:
    case COLLISION_UP_RIGHT:
    case COLLISION_DOWN_RIGHT:
      v_y = -v_y;
      v_x = -v_x;
      break;
    case NO_COLLISION:
      exit(0);
    }
  }

  return 0;
}
