//--------------------------------------------------------------
// Copyright (C) 2004 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: field.c,v 1.2 2004/09/16 23:42:01 schaum Exp $
//
//--------------------------------------------------------------
#include "field.h"

#define PADDLE_HEIGHT (display_lines()/8)

void send_paddle_height() {
  // to paddle 1
  send_byte_2(PADDLE_HEIGHT);
  send_byte_2(display_lines());
  send_byte_2(display_cols());

  // to paddle 2
  send_byte_3(PADDLE_HEIGHT);
  send_byte_3(display_lines());
  send_byte_3(display_cols());

  // to ball
  send_byte(display_lines());
  send_byte(display_cols());

}

void get_ball_position(int *ball_x, int *ball_y, int *v_x, int *v_y, int *ct) {
  *ball_x         = recv_byte();
  *ball_y         = recv_byte();
  *v_x            = recv_byte();
  *v_y            = recv_byte();
  *ct             = recv_byte();
}

int no_new_ball() {
  return (poll_byte() == 0);
}

void update_paddle1_pos(int *p) {
  if (poll_byte_2())
    *p = recv_byte_2();
}

int get_paddle1_pos() {
  return recv_byte_2();
}

void update_paddle2_pos(int *p) {
  if (poll_byte_3())
    *p = recv_byte_3();
}

int get_paddle2_pos() {
  return recv_byte_3();
}

void send_collision(int ball_x, int ball_y, int collision_type) {
  send_byte(ball_x);
  send_byte(ball_y);
  send_byte(collision_type);
}

void clear_display() {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  *disp = 0x80000000;
  *disp = 0;
}

void update_display() {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  *disp = 0x80000000 | (3 << 24);
  *disp = 0;
  *disp = 0x80000000 | (1 << 24);
  *disp = 0;
}

int display_cols() {
  volatile unsigned int *disp = (volatile unsigned int *) 0x8000000C;
  return *disp;
}

int display_lines() {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000010;
  return *disp;
}

void clear_ball_pos(int x, int y) {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  int v = (x & 0xff) | ((y & 0xff) << 8) | ((' ' & 0xff) << 16) | (2 << 24);

  *disp = v | 0x80000000;
  *disp = v;
}

void display_ball_pos(int x, int y) {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  int v = (x & 0xff) | ((y & 0xff) << 8) | (('0' & 0xff) << 16) | (2 << 24);

  *disp = v | 0x80000000;
  *disp = v;
}

void display_bounce() {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  *disp = 0x80000000 | (4 << 24);
  *disp = 0;
}

void display_paddle_pos(int x, int y) {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  int i;

  for (i = 0; i < PADDLE_HEIGHT; i++) {
    int v = (x & 0xff) | ((y+i & 0xff) << 8) | (('X' & 0xff) << 16) | (2 << 24);
    *disp = v | 0x80000000;
    *disp = v;
    v = (x & 0xff) | ((y-i & 0xff) << 8) | (('X' & 0xff) << 16) | (2 << 24);
    *disp = v | 0x80000000;
    *disp = v;
  }
}

void clear_paddle_pos(int x, int y) {
  volatile unsigned int *disp = (volatile unsigned int *) 0x80000008;
  int i;

  for (i = 0; i < PADDLE_HEIGHT; i++) {
    int v = (x & 0xff) | ((y+i & 0xff) << 8) | ((' ' & 0xff) << 16) | (2 << 24);
    *disp = v | 0x80000000;
    *disp = v;
    v = (x & 0xff) | ((y-i & 0xff) << 8) | ((' ' & 0xff) << 16) | (2 << 24);
    *disp = v | 0x80000000;
    *disp = v;
  }
}

void move_ball(int *ball_x, int *ball_y, int *v_x, int *v_y) {
  clear_ball_pos(*ball_x, *ball_y);

  *ball_x += *v_x;
  *ball_y += *v_y;

  display_ball_pos(*ball_x, *ball_y);
}

int collided(int ball_x, int v_x, int ball_y, int v_y) {

  int collision_updown    = NO_COLLISION;
  int collision_leftright = NO_COLLISION;

  if ((ball_y <= (display_lines()-1)) && ((ball_y + v_y) > (display_lines()-1))) 
    collision_updown = COLLISION_DOWN;
  if ((ball_y >= 0)                   && ((ball_y + v_y) < 0))
    collision_updown = COLLISION_UP;
  if ((ball_x <= (display_cols()-1))  && ((ball_x + v_x) > (display_cols()-1)))
    return BALL_LOST;
  if ((ball_x >= 0)                   && ((ball_x + v_x) < 0))
    return BALL_LOST;
  
  return collision_updown;
}

int update_delay() {
  int i;
  for (i=0; i<1000; i++) ;
}

int main() {
  int ball_x, ball_y, ct;
  int new_ball_x, new_ball_y;
  int v_x, v_y;
  int new_v_x, new_v_y;
  int collision_type;
  int i;

  int paddle_1_pos = 5;
  int paddle_2_pos = 5;

  data_outp   = (volatile unsigned int *) 0x80000004;
  data_inp    = (volatile unsigned int *) 0x80000000;
  data_outp_2 = (volatile unsigned int *) 0x80000018;
  data_inp_2  = (volatile unsigned int *) 0x80000014;
  data_outp_3 = (volatile unsigned int *) 0x80000020;
  data_inp_3  = (volatile unsigned int *) 0x8000001C;

  // initialization
  send_paddle_height();
  get_ball_position(&ball_x, &ball_y, &v_x, &v_y, &ct);
  paddle_1_pos = get_paddle1_pos();
  paddle_2_pos = get_paddle2_pos();

  display_paddle_pos(PADDLE_FROM_EDGE, paddle_1_pos);

  while (1) {

    get_ball_position(&new_ball_x, &new_ball_y, &new_v_x, &new_v_y, &ct);

    if (ct != BALL_LOST) {
      // must make sure we have continous
      if (((ball_x - new_ball_x) >= v_x) && (((ball_y - new_ball_y) >= v_y))) {
	// anticausal situation detected - we have moved the ball to
	// ball_x and ball_y already but a collision should have happened at
	// new_ball_x and new_ball_y
	// In this case, simply move the ball to new_ball_x and new_ball_y with
	// new velocity
	clear_ball_pos(ball_x, ball_y);
	ball_x = new_ball_x;
	ball_y = new_ball_y;      
      }
      
      // this loop makes sure we have a continous movement of the ball,
      // independent how soon collisions are announced 
      while ((new_ball_x != ball_x) || (new_ball_y != ball_y)) {
	move_ball(&ball_x, &ball_y, &v_x, &v_y);
	update_delay();
	update_display();
      }
      
    } else {
      clear_ball_pos(ball_x, ball_y);
      ball_x = new_ball_x;
      ball_y = new_ball_y;
    }

    v_x = new_v_x;
    v_y = new_v_y;

    do {
      move_ball(&ball_x, &ball_y, &v_x, &v_y);

      clear_paddle_pos  (  PADDLE_FROM_EDGE, paddle_1_pos);
      update_paddle1_pos( &paddle_1_pos);
      display_paddle_pos(  PADDLE_FROM_EDGE, paddle_1_pos);

      clear_paddle_pos  (  display_cols() - 1 - PADDLE_FROM_EDGE, paddle_2_pos);
      update_paddle2_pos( &paddle_2_pos);
      display_paddle_pos(  display_cols() - 1 - PADDLE_FROM_EDGE, paddle_2_pos);

      update_delay();
      update_display() ;
      collision_type = collided(ball_x, v_x, ball_y, v_y);

    } while ((collision_type == NO_COLLISION) && no_new_ball());

    if ((collision_type != NO_COLLISION)) {
      display_bounce();
      send_collision(ball_x, ball_y, collision_type);
    }
  }

  return 0;
}
