//--------------------------------------------------------------
// Copyright (C) 2004 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: paddle1.c,v 1.1 2004/03/10 06:28:02 schaum Exp $
//
//--------------------------------------------------------------

#include "field.h"

int paddle_height;
int field_y;
int field_x;

void get_paddle_height() {
  paddle_height = recv_byte_2();
  field_y       = recv_byte_2();
  field_x       = recv_byte_2();
}

void send_paddle_pos(int p) {
  send_byte_2(p);
}

void send_collision(int ball_x, int ball_y, int collision_type) {
  send_byte(ball_x);
  send_byte(ball_y);
  send_byte(collision_type);
}

void get_ball_position(int *ball_x, int *ball_y, int *v_x, int *v_y, int *ct) {
  *ball_x         = recv_byte();
  *ball_y         = recv_byte();
  *v_x            = recv_byte();
  *v_y            = recv_byte();
  *ct             = recv_byte();
}

int no_new_ball() {
  return (poll_byte() == 0);
}

void move_ball(int *ball_x, int *ball_y, int *v_x, int *v_y) {
  if ((*ball_x >= PADDLE_FROM_EDGE) && (*v_x <= 0)) {
    // collision is possible
    *ball_y = *ball_y + *v_y;
    *ball_x = *ball_x + *v_x;
  }
}

int collided(int ball_x, int v_x, int ball_y, int paddle_position) {
  if ((ball_x > PADDLE_FROM_EDGE) && ((ball_x + v_x) <= PADDLE_FROM_EDGE)) {
    if ((ball_y <= (paddle_position + paddle_height)) && 
	(ball_y >= (paddle_position - paddle_height))) {
      return COLLISION_LEFT;
    }
  }
  return NO_COLLISION;
}

int move_paddle(int ball_y, int v_y, int paddle_position) {
  if (ball_y < paddle_position)
    if (paddle_position > paddle_height)
      return paddle_position - 1;
  if (ball_y > paddle_position)
    if (paddle_position < (field_y - paddle_height))
      return paddle_position + 1;
  return paddle_position;
}

int main() {

  data_outp   = (volatile unsigned int *) 0x80000004;
  data_inp    = (volatile unsigned int *) 0x80000000;
  data_outp_2 = (volatile unsigned int *) 0x80000008;
  data_inp_2  = (volatile unsigned int *) 0x8000000C;

  int bx, by, vx, vy, ct, i;
  int new_bx, new_by, new_vx, new_vy;
  int paddle_position = 10, new_paddle_position;
  int collision_type;

  // initialization
  get_paddle_height();
  get_ball_position(&bx, &by, &vx, &vy, &ct);

  send_paddle_pos(paddle_position);

  while (1) {
    get_ball_position(&bx, &by, &vx, &vy, &ct);

    do {
      move_ball(&bx, &by, &vx, &vy);
      new_paddle_position = move_paddle(by, vy, paddle_position);

      if (new_paddle_position != paddle_position) {
	send_paddle_pos(new_paddle_position);
      }
      paddle_position = new_paddle_position;

      collision_type = collided(bx, vx, by, paddle_position);

    } while ((collision_type == NO_COLLISION) && no_new_ball());

    if ((collision_type != NO_COLLISION)) {
      send_collision(bx, by, collision_type);
    } 

  }

  return 0;
}
